from typing import List

from qtpy import QtCore, QtWidgets

from activity_browser import application
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.bwutils.strategies import relink_activity_exchanges
from activity_browser.mod import bw2data as bd
from activity_browser.ui.icons import qicons


class ActivityRelink(ABAction):
    """
    ABAction to relink the exchanges of an activity to exchanges from another database.

    This action only uses the first key from activity_keys
    """

    icon = qicons.edit
    text = "Relink the activity exchanges"

    @staticmethod
    @exception_dialogs
    def run(activity_keys: List[tuple]):
        # this action only uses the first key supplied to activity_keys
        key = activity_keys[0]

        # extract the brightway database and activity
        db = bd.Database(key[0])
        activity = bd.get_activity(key)

        # find the dependents for the database and construct the alternatives in tuple format
        depends = db.find_dependents()
        options = [(depend, list(bd.databases)) for depend in depends]

        # present the alternatives to the user in a linking dialog
        dialog = ActivityLinkingDialog.relink_sqlite(
            activity["name"], options, application.main_window
        )

        # return if the user cancels
        if dialog.exec_() == ActivityLinkingDialog.Rejected:
            return

        # relinking will take some time, set WaitCursor
        QtWidgets.QApplication.setOverrideCursor(QtCore.Qt.WaitCursor)

        # use the relink_activity_exchanges strategy to relink the exchanges of the activity
        relinking_results = {}
        for old, new in dialog.relink.items():
            other = bd.Database(new)
            failed, succeeded, examples = relink_activity_exchanges(
                activity, old, other
            )
            relinking_results[f"{old} --> {other.name}"] = (failed, succeeded)

        # restore normal cursor
        QtWidgets.QApplication.restoreOverrideCursor()

        # if any relinks failed present them to the user
        if failed > 0:
            relinking_dialog = ActivityLinkingResultsDialog.present_relinking_results(
                application.main_window, relinking_results, examples
            )
            relinking_dialog.exec_()


class ActivityLinkingDialog(QtWidgets.QDialog):
    """
    Displays the possible databases for relinking the exchanges for a given activity
    """

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Activity linking")

        self.db_label = QtWidgets.QLabel()
        self.label_choices = []
        self.grid_box = QtWidgets.QGroupBox("Database links:")
        self.grid = QtWidgets.QGridLayout()
        self.grid_box.setLayout(self.grid)

        self.buttons = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel,
        )
        self.buttons.accepted.connect(self.accept)
        self.buttons.rejected.connect(self.reject)

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.db_label)
        layout.addWidget(self.grid_box)
        layout.addWidget(self.buttons)
        self.setLayout(layout)

    @property
    def relink(self) -> dict:
        """Returns a dictionary of str -> str key/values, showing which keys
        should be linked to which values.

        Only returns key/value pairs if they differ.
        """
        return {
            label.text(): combo.currentText()
            for label, combo in self.label_choices
            if label.text() != combo.currentText()
        }

    @property
    def links(self) -> dict:
        """Returns a dictionary of str -> str key/values, showing which keys
        should be linked to which values.
        """
        return {
            label.text(): combo.currentText() for label, combo in self.label_choices
        }

    @classmethod
    def construct_dialog(
        cls,
        label: str,
        options: list,
        parent: QtWidgets.QWidget = None,
    ) -> "ActivityLinkingDialog":
        obj = cls(parent)
        obj.db_label.setText(label)
        # Start at 1 because row 0 is taken up by the db_label
        for i, item in enumerate(options):
            label = QtWidgets.QLabel(item[0])
            combo = QtWidgets.QComboBox()
            combo.addItems(item[1])
            combo.setCurrentText(item[0])
            obj.label_choices.append((label, combo))
            obj.grid.addWidget(label, i, 0, 1, 2)
            obj.grid.addWidget(combo, i, 2, 1, 2)
        obj.updateGeometry()
        return obj

    @classmethod
    def relink_sqlite(
        cls, act: str, options: list, parent=None
    ) -> "ActivityLinkingDialog":
        label = "Relinking exchanges from activity '{}'.".format(act)
        return cls.construct_dialog(label, options, parent)


class ActivityLinkingResultsDialog(QtWidgets.QDialog):
    """
    Provides a summary from a relinking of activity exchanges for the relinking of a
    single activity.
    A simple design layout based on the DatabaseLinkingResultsDialog
    """

    def __init__(self, parent=None):
        super().__init__(parent)

        self.setWindowTitle("Relinking database results")

        button = QtWidgets.QDialogButtonBox.Ok
        self.buttonBox = QtWidgets.QDialogButtonBox(button)
        self.buttonBox.accepted.connect(self.accept)
        self.databases_relinked = QtWidgets.QVBoxLayout()

        self.activityToOpen = set()

        self.exchangesUnlinked = QtWidgets.QVBoxLayout()

        self.layout = QtWidgets.QVBoxLayout()
        self.layout.addLayout(self.databases_relinked)
        self.layout.addLayout(self.exchangesUnlinked)
        self.layout.addWidget(self.buttonBox)
        self.setLayout(self.layout)

    @classmethod
    def construct_results_dialog(
        cls,
        parent: QtWidgets.QWidget = None,
        link_results: dict = None,
        unlinked_exchanges: dict = None,
    ) -> "ActivityLinkingResultsDialog":
        from activity_browser import actions

        obj = cls(parent)
        for k, results in link_results.items():
            obj.databases_relinked.addWidget(
                QtWidgets.QLabel(f"{k} = {results[1]} successfully linked")
            )
            obj.databases_relinked.addWidget(
                QtWidgets.QLabel(f"{k} = {results[0]} flows failed to link")
            )

        obj.exchangesUnlinked.addWidget(
            QtWidgets.QLabel("Up to 5 unlinked exchanges (click to open)")
        )
        for act, key in unlinked_exchanges.items():
            button = QtWidgets.QPushButton(act.as_dict()["name"])
            button.clicked.connect(
                lambda: actions.ActivityOpen.run([act.key])
            )
            obj.exchangesUnlinked.addWidget(button)
        obj.updateGeometry()

        return obj

    @classmethod
    def present_relinking_results(
        cls,
        parent: QtWidgets.QWidget = None,
        link_results: dict = None,
        unlinked_exchanges: dict = None,
    ) -> "ActivityLinkingResultsDialog":
        return cls.construct_results_dialog(parent, link_results, unlinked_exchanges)

    def select_activity_to_open(self, actvty: tuple) -> None:
        if actvty in self.activityToOpen:
            self.activityToOpen.discard(actvty)
        self.activityToOpen.add(actvty)

    def open_activity(self):
        return self.activityToOpen




# -*- coding: utf-8 -*
import abc

import numpy as np
from bw2data.parameters import ParameterBase
from bw2data.proxies import ExchangeProxyBase
from stats_arrays import UncertaintyBase, UndefinedUncertainty
from stats_arrays import uncertainty_choices as uc

EMPTY_UNCERTAINTY = {
    "uncertainty type": UndefinedUncertainty.id,
    "loc": np.NaN,
    "scale": np.NaN,
    "shape": np.NaN,
    "minimum": np.NaN,
    "maximum": np.NaN,
    "negative": False,
}


class BaseUncertaintyInterface(abc.ABC):
    __slots__ = ["_data"]
    KEYS = {
        "uncertainty type",
        "loc",
        "scale",
        "shape",
        "minimum",
        "maximum",
        "negative",
        "pedigree",
    }
    data_type = ""

    def __init__(self, unc_obj):
        self._data = unc_obj

    @property
    def data(self):
        return self._data  # pragma: no cover

    @property
    @abc.abstractmethod
    def amount(self) -> float:
        pass  # pragma: no cover

    @property
    @abc.abstractmethod
    def uncertainty_type(self) -> UncertaintyBase:
        pass  # pragma: no cover

    @property
    @abc.abstractmethod
    def uncertainty(self) -> dict:
        pass  # pragma: no cover


class ExchangeUncertaintyInterface(BaseUncertaintyInterface):
    """Many kinds of exchanges use uncertainty to describe how 'correct' the
    data is which makes up the amount of the exchange.
    """

    _data: ExchangeProxyBase
    data_type = "exchange"

    @property
    def amount(self) -> float:
        return self._data.amount

    @property
    def uncertainty_type(self) -> UncertaintyBase:
        return self._data.uncertainty_type

    @property
    def uncertainty(self) -> dict:
        return self._data.uncertainty


class ParameterUncertaintyInterface(BaseUncertaintyInterface):
    """All levels of parameters can describe their amounts with uncertainty."""

    _data: ParameterBase
    data_type = "parameter"

    @property
    def amount(self) -> float:
        return self._data.amount

    @property
    def uncertainty_type(self) -> UncertaintyBase:
        return uc[self._data.data.get("uncertainty type", 0)]

    @property
    def uncertainty(self) -> dict:
        return {k: self._data.data.get(k) for k in self.KEYS if k in self._data.data}


class CFUncertaintyInterface(BaseUncertaintyInterface):
    """The characterization factors (CFs) of an impact assessment method can also
    contain uncertainty.

    This is however not certain (ha), as the CF is made up out of a flow key
    and either an amount (float) or the uncertainty values + an amount (dict).

    See the ``Method`` and ``ProcessedDataStore`` classes in the bw2data library.
    """

    _data: tuple
    data_type = "cf"

    @property
    def is_uncertain(self) -> bool:
        return isinstance(self._data[1], dict)

    @property
    def amount(self) -> float:
        return self._data[1]["amount"] if self.is_uncertain else self._data[1]

    @property
    def uncertainty_type(self) -> UncertaintyBase:
        if not self.is_uncertain:
            return UndefinedUncertainty
        return uc[self._data[1].get("uncertainty type", 0)]

    @property
    def uncertainty(self) -> dict:
        if not self.is_uncertain:
            return {}
        return {k: self._data[1].get(k) for k in self.KEYS if k in self._data[1]}


def get_uncertainty_interface(data: object) -> BaseUncertaintyInterface:
    if isinstance(data, ExchangeProxyBase):
        return ExchangeUncertaintyInterface(data)
    elif isinstance(data, ParameterBase):
        return ParameterUncertaintyInterface(data)
    elif isinstance(data, tuple):
        return CFUncertaintyInterface(data)
    else:
        raise TypeError(
            "No uncertainty interface exists for object type {}".format(type(data))
        )



from os import name
from typing import List
from logging import getLogger

from qtpy import QtWidgets

from activity_browser import application
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.mod import bw2data as bd
from activity_browser.ui.icons import qicons

log = getLogger(__name__)


class MethodDelete(ABAction):
    """
    Delete one or more impact assessment methods (impact categories).

    Flow:
    - Confirm with the user.
    - Deregister all selected methods from Brightway.
    - Update all Brightway calculation setups by removing the deleted methods from their
      'ia' list (if present), and serialize the updated setups.

    Notes:
    - Calculation setups can store method identifiers either as tuples (recommended) or
      sometimes as strings for single-level names; the cleanup accounts for both.
    """

    icon = qicons.delete
    text = "Delete Impact Category"

    @staticmethod
    @exception_dialogs
    def run(methods: List[tuple]):
        # check whether we're dealing with a leaf or node. If it's a node, select all underlying methods for deletion
        all_methods = [bd.Method(method) for method in methods]

        if len(all_methods) == 1:
            warning_text = f"Are you sure you want to delete this method?\n\n{methods[0]}"
        else:
            warning_text = f"Are you sure you want to delete {len(all_methods)} methods?"

        # warn the user about the pending deletion
        warning = QtWidgets.QMessageBox.warning(
            application.main_window,
            "Deleting Method",
            warning_text,
            QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
            QtWidgets.QMessageBox.No,
        )
        # return if the users cancels
        if warning == QtWidgets.QMessageBox.No:
            return

        # collect names for calculation setup cleanup
        to_remove = {m.name for m in all_methods}

        # delete all methods by deregistering them
        for method in all_methods:
            method.deregister()
            log.info(f"Deleted method {method.name}")

        # remove deleted methods from all calculation setups
        MethodDelete.remove_methods_from_calculation_setups(to_remove)

    @staticmethod
    def remove_methods_from_calculation_setups(method_names: set[tuple]) -> None:
        """
        Remove given method names from all calculation setups' 'ia' lists and serialize.
        """
        try:
            changed_any = False

            for cs_name, cs in bd.calculation_setups.items():
                ia = cs.get("ia", [])
                
                for name in method_names:
                    if name not in ia:
                        continue  # name not present, skip
                    
                    ia.remove(name)
                    changed_any = True
                    
                    log.info(
                        f"Updated calculation setup '{cs_name}': removed impact category {name}"
                    )

            
            if changed_any:
                bd.calculation_setups.serialize()
        except Exception:
            log.exception("Failed to update calculation setups after method rename")



# -*- coding: utf-8 -*-
import math
from logging import getLogger

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg
from matplotlib.figure import Figure
from qtpy import QtWidgets

from bw2data import methods
from activity_browser.utils import savefilepath
from activity_browser.bwutils.commontasks import wrap_text


log = getLogger(__name__)

# todo: sizing of the figures needs to be improved and systematized...
# todo: Bokeh is a potential alternative as it allows interactive visualizations,
#  but this issue needs to be resolved first: https://github.com/bokeh/bokeh/issues/8169


class Plot(QtWidgets.QWidget):
    ALL_FILTER = "All Files (*.*)"
    PNG_FILTER = "PNG (*.png)"
    SVG_FILTER = "SVG (*.svg)"

    def __init__(self, parent=None):
        super().__init__(parent)
        # create figure, canvas, and axis
        # self.figure = Figure(tight_layout=True)
        self.figure = Figure(constrained_layout=True)
        self.canvas = FigureCanvasQTAgg(self.figure)
        self.canvas.setMinimumHeight(0)

        self.canvas.destroyed.connect(self.check)

        self.ax = self.figure.add_subplot(111)  # create an axis
        self.plot_name = "Figure"

        # set the layout
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.canvas)
        self.setLayout(layout)
        self.setSizePolicy(
            QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding
        )
        self.updateGeometry()

    def check(self):
        print("WHY DELETE")

    def plot(self, *args, **kwargs):
        raise NotImplementedError

    def reset_plot(self) -> None:
        self.figure.clf()
        self.ax = self.figure.add_subplot(111)

    def get_canvas_size_in_inches(self):
        # print("Canvas size:", self.canvas.get_width_height())
        return tuple(x / self.figure.dpi for x in self.canvas.get_width_height())

    def to_png(self):
        """Export to .png format."""
        filepath = savefilepath(
            default_file_name=self.plot_name, file_filter=self.PNG_FILTER
        )
        if filepath:
            if not filepath.endswith(".png"):
                filepath += ".png"
            self.figure.savefig(filepath)

    def to_svg(self):
        """Export to .svg format."""
        filepath = savefilepath(
            default_file_name=self.plot_name, file_filter=self.SVG_FILTER
        )
        if filepath:
            if not filepath.endswith(".svg"):
                filepath += ".svg"
            self.figure.savefig(filepath)


class LCAResultsBarChart(Plot):
    """ " Generate a bar chart comparing the absolute LCA scores of the products"""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.plot_name = "LCA scores"

    def plot(self, df: pd.DataFrame, method: tuple, labels: list):
        self.reset_plot()
        height_inches, width_inches = self.get_canvas_size_in_inches()
        self.figure.set_size_inches(height_inches, width_inches)

        # https://github.com/LCA-ActivityBrowser/activity-browser/issues/489
        df.index = pd.Index(labels)  # Replace index of tuples
        show_legend = df.shape[1] != 1  # Do not show the legend for 1 column
        df.plot.barh(ax=self.ax, legend=show_legend)
        self.ax.invert_yaxis()

        # labels
        self.ax.set_yticks(np.arange(len(labels)))
        self.ax.set_xlabel(methods[method].get("unit"))
        self.ax.set_title(", ".join([m for m in method]))
        # self.ax.set_yticklabels(labels, minor=False)

        # grid
        self.ax.grid(which="major", axis="x", color="grey", linestyle="dashed")
        self.ax.set_axisbelow(True)  # puts gridlines behind bars

        # draw
        self.canvas.draw()


class LCAResultsPlot(Plot):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.plot_name = "LCA heatmap"

    def plot(self, df: pd.DataFrame, invert_plot: bool = False):
        """Plot a heatmap grid of the different impact categories and reference flows."""
        # need to clear the figure and add axis again
        # because of the colorbar which does not get removed by the ax.clear()
        self.reset_plot()

        dfp = df.copy()
        dfp.index = dfp["index"]
        dfp.drop(
            dfp.select_dtypes(["object"]), axis=1, inplace=True
        )  # get rid of all non-numeric columns (metadata)
        if "amount" in dfp.columns:
            dfp.drop(["amount"], axis=1, inplace=True)  # Drop the 'amount' col
        if "Score" in dfp.index:
            dfp.drop("Score", inplace=True)

        # avoid figures getting too large horizontally
        dfp.index = [wrap_text(i, max_length=40) for i in dfp.index]
        dfp.columns = [wrap_text(i, max_length=20) for i in dfp.columns]
        prop = dfp.divide(dfp.abs().max(axis=0)).multiply(100)
        dfp.replace(np.nan, 0, inplace=True)
        if invert_plot:
            dfp = dfp.T
            prop = prop.T

        # set different color palette depending on whether all values are positive or not
        if (
            dfp.min(axis=None) < 0 and dfp.max(axis=None) > 0
        ):  # has both negative AND positive values
            cmap = sns.color_palette("vlag_r", as_cmap=True)
        else:  # has only positive OR negative values
            cmap = sns.color_palette("Blues", as_cmap=True)

        sns.heatmap(
            prop,
            ax=self.ax,
            cmap=cmap,
            annot=dfp,
            linewidths=0.05,
            annot_kws={
                "size": 11 if dfp.shape[1] <= 8 else 9,
                "rotation": 0 if dfp.shape[1] <= 8 else 60,
            },
            cbar_kws={"format": "%.0f%%"},
        )
        self.ax.tick_params(labelsize=8)
        if dfp.shape[1] > 5:
            self.ax.set_xticklabels(self.ax.get_xticklabels(), rotation="vertical")
        self.ax.set_yticklabels(self.ax.get_yticklabels(), rotation="horizontal")

        # refresh canvas
        size_inches = (2 + dfp.shape[0] * 0.5, 4 + dfp.shape[0] * 0.55)
        self.figure.set_size_inches(self.get_canvas_size_in_inches()[0], size_inches[1])
        size_pixels = self.figure.get_size_inches() * self.figure.dpi
        self.setMinimumHeight(size_pixels[1])

        self.canvas.draw()


class ContributionPlot(Plot):
    MAX_LEGEND = 30

    def __init__(self, parent=None):
        super().__init__(parent)
        self.plot_name = "Contributions"
        self.parent = parent

    def plot(self, df: pd.DataFrame, unit: str = None):
        """Plot a horizontal stacked bar chart of contributions,
        add 'total' marker if both positive and negative results are present."""
        dfp = df.copy()
        dfp = dfp.iloc[:, ::-1]  # reverse column names so they align with calculation setup and rest of results

        dfp.index = dfp["index"]
        dfp.drop(
            dfp.select_dtypes(["object"]), axis=1, inplace=True
        )  # get rid of all non-numeric columns (metadata)
        if "Score" in dfp.index:
            dfp.drop("Score", inplace=True)
        if "id" in dfp:
            dfp.drop(columns=["id"], inplace=True)
        # drop rows if all values are 0 except for "Rest (+)" and "Rest (-)"
        rows_to_drop = dfp.index[(dfp == 0).all(axis=1) & ~dfp.index.isin(["Rest (+)", "Rest (-)"])]
        # Drop those rows
        dfp = dfp.drop(rows_to_drop)

        self.ax.clear()
        canvas_width_inches, canvas_height_inches = self.get_canvas_size_in_inches()
        optimal_height_inches = 4 + dfp.shape[1] * 0.55
        # print('Optimal Contribution plot height:', optimal_height_inches)
        self.figure.set_size_inches(canvas_width_inches, optimal_height_inches)

        # avoid figures getting too large horizontally
        dfp.index = pd.Index([wrap_text(str(i), max_length=40) for i in dfp.index])
        dfp.columns = pd.Index([wrap_text(i, max_length=40) for i in dfp.columns])
        # Strip invalid characters from the ends of row/column headers
        dfp.index = dfp.index.str.strip("_ \n\t")
        dfp.columns = dfp.columns.str.strip("_ \n\t")

        # set colormap to use
        items = dfp.shape[0]  # how many contribution items
        # skip grey and black at start/end of cmap
        cmap = plt.cm.nipy_spectral_r(np.linspace(0, 1, items + 2))[1:-1]
        colors = {item: color for item, color in zip(dfp.index, cmap)}
        # overwrite rest values to grey
        colors["Rest (+)"] = [0.8, 0.8, 0.8, 1.]
        colors["Rest (-)"] = [0.8, 0.8, 0.8, 1.]

        dfp.T.plot.barh(
            stacked=True,
            color=colors,
            ax=self.ax,
            legend=False if dfp.shape[0] >= self.MAX_LEGEND else True,
        )
        self.ax.tick_params(labelsize=8)
        if unit:
            self.ax.set_xlabel(unit)

        # show legend if not too many items
        if not dfp.shape[0] >= self.MAX_LEGEND:
            plt.rc("legend", **{"fontsize": 8})
            ncols = math.ceil(dfp.shape[0] * 0.6 / optimal_height_inches)
            # print('Ncols:', ncols, dfp.shape[0] * 0.55, optimal_height_inches)
            self.ax.legend(loc="center left", bbox_to_anchor=(1, 0.5), ncol=ncols)

        # grid
        self.ax.grid(which="major", axis="x", color="grey", linestyle="dashed")
        self.ax.set_axisbelow(True)  # puts gridlines behind bars
        # make the zero line more present
        grid = self.ax.get_xgridlines()
        # get the 0 line from all gridlines
        label_pos = [i for i, label in enumerate(self.ax.get_xticklabels()) if label.get_position()[0] == 0.0]
        if len(label_pos) > 0:
            zero_line = grid[label_pos[0]]
            zero_line.set_color("black")
            zero_line.set_linestyle("solid")

        # total marker when enabled and both negative and positive results are present in a column
        if self.parent.score_marker:
            marker_size = max(min(150 / dfp.shape[1], 35), 10)  # set marker size dynamic between 10 - 35
            for i, col in enumerate(dfp):
                total = np.sum(dfp[col])
                abs_total = np.sum(np.abs(dfp[col]))
                if abs(total) != abs_total:
                    self.ax.plot(total, i,
                                 markersize=marker_size, marker="d", fillstyle="left",
                                 markerfacecolor="black", markerfacecoloralt="grey", markeredgecolor="white")

        # TODO review: remove or enable

        # refresh canvas
        # size_inches = (2 + dfp.shape[0] * 0.5, 4 + dfp.shape[1] * 0.55)
        # self.figure.set_size_inches(self.get_canvas_size_in_inches()[0], size_inches[1])

        size_pixels = self.figure.get_size_inches() * self.figure.dpi
        self.setMinimumHeight(size_pixels[1])
        self.canvas.draw()


class CorrelationPlot(Plot):
    def __init__(self, parent=None):
        super().__init__(parent)
        sns.set(style="darkgrid")

    def plot(self, df: pd.DataFrame):
        """Plot a heatmap of correlations between different reference flows."""
        # need to clear the figure and add axis again
        # because of the colorbar which does not get removed by the ax.clear()
        self.reset_plot()
        canvas_size = self.canvas.get_width_height()
        # print("Canvas size:", canvas_size)
        size = (4 + df.shape[1] * 0.3, 4 + df.shape[1] * 0.3)
        self.figure.set_size_inches(size[0], size[1])

        corr = df.corr()
        # Generate a mask for the upper triangle
        mask = np.zeros_like(corr, dtype=bool)
        mask[np.triu_indices_from(mask)] = True
        # Draw the heatmap with the mask and correct aspect ratio
        vmax = np.abs(corr.values[~mask]).max()
        # vmax = np.abs(corr).max()
        sns.heatmap(
            corr,
            mask=mask,
            cmap=plt.cm.PuOr,
            vmin=-vmax,
            vmax=vmax,
            square=True,
            linecolor="lightgray",
            linewidths=1,
            ax=self.ax,
        )

        df_lte8_cols = df.shape[1] <= 8
        for i in range(len(corr)):
            self.ax.text(
                i + 0.5,
                i + 0.5,
                corr.columns[i],
                ha="center",
                va="center",
                rotation=0 if df_lte8_cols else 45,
                size=11 if df_lte8_cols else 9,
            )
            for j in range(i + 1, len(corr)):
                s = "{:.3f}".format(corr.values[i, j])
                self.ax.text(
                    j + 0.5,
                    i + 0.5,
                    s,
                    ha="center",
                    va="center",
                    rotation=0 if df_lte8_cols else 45,
                    size=11 if df_lte8_cols else 9,
                )
        self.ax.axis("off")

        # refresh canvas
        size_pixels = self.figure.get_size_inches() * self.figure.dpi
        self.setMinimumHeight(size_pixels[1])
        self.canvas.draw()


class MonteCarloPlot(Plot):
    """Monte Carlo plot."""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.plot_name = "Monte Carlo"

    def plot(self, df: pd.DataFrame, method: tuple):
        self.ax.clear()

        for col in df.columns:
            color = self.ax._get_lines.get_next_color()
            df[col].hist(
                ax=self.ax,
                figure=self.figure,
                label=col,
                density=True,
                color=color,
                alpha=0.5,
            )  # , histtype="step")
            # self.ax.axvline(df[col].median(), color=color)
            self.ax.axvline(df[col].mean(), color=color)

        self.ax.set_xlabel(methods[method]["unit"])
        self.ax.set_ylabel("Probability")
        self.ax.legend(
            loc="upper center",
            bbox_to_anchor=(0.5, -0.07),
        )  # ncol=2

        # lconfi, upconfi =mc['statistics']['interval'][0], mc['statistics']['interval'][1]

        self.canvas.draw()



from typing import List

from qtpy import QtWidgets

from activity_browser import application
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.mod import bw2data as bd
from activity_browser.ui.icons import qicons


class CFNew(ABAction):
    """
    ABAction to add a new characterization flow to a method through one or more elementary-flow keys.
    """

    icon = qicons.add
    text = "New characterization factor"

    @staticmethod
    @exception_dialogs
    def run(method_name: tuple, keys: List[tuple]):
        # load old cf's from the Method
        method_dict = {cf[0]: cf[1] for cf in bd.Method(method_name).load()}

        # use only the keys that don't already exist within the method
        unique_keys = [key for key in keys if key not in method_dict]

        # if there are non-unique keys warn the user that these won't be added
        if len(unique_keys) < len(keys):
            QtWidgets.QMessageBox.warning(
                application.main_window,
                "Duplicate characterization factors",
                "One or more of these elementary flows already exist within this method. Duplicate flows will not be "
                "added",
            )

        # return if there are no new keys
        if not unique_keys:
            return

        # add the new keys to the method dictionary
        for key in unique_keys:
            method_dict[key] = 0.0

        # write the updated dict to the method
        bd.Method(method_name).write(list(method_dict.items()))



from qtpy import QtWidgets, QtCore
from activity_browser import actions
from activity_browser.ui import widgets


class ImpactCategoryHeader(QtWidgets.QWidget):

    def __init__(self, parent: QtWidgets.QWidget):
        """
        Initializes the ImpactCategoryHeader widget with a stack layout
        that switches between editable and view-only headers.

        Args:
            parent (QtWidgets.QWidget): The parent widget.
        """
        super().__init__(parent)
        self.impact_category = parent.impact_category

        # Set size policy to only take needed vertical space
        self.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Expanding,
            QtWidgets.QSizePolicy.Policy.Maximum
        )

        # Create stack layout to hold both header types
        self.stack = QtWidgets.QStackedLayout()
        self.stack.setContentsMargins(0, 0, 0, 0)
        self.stack.setSizeConstraint(QtWidgets.QLayout.SizeConstraint.SetMinimumSize)
        
        # Create both header widgets
        self.view_only_header = ViewOnlyHeader(self)
        self.editable_header = EditableHeader(self)
        
        # Add headers to stack
        self.stack.addWidget(self.view_only_header)  # Index 0
        self.stack.addWidget(self.editable_header)   # Index 1
        
        self.setLayout(self.stack)

    def sync(self):
        """
        Synchronizes the widget with the current state of the impact category.
        Switches between editable and view-only headers based on edit mode.
        """
        self.impact_category = self.parent().impact_category
        
        # Update both headers with current data
        self.view_only_header.sync()
        self.editable_header.sync()
        
        # Switch to appropriate header based on edit mode
        if self.parent().is_editable:
            self.stack.setCurrentIndex(1)  # Show editable header
        else:
            self.stack.setCurrentIndex(0)  # Show view-only header

    def on_editable_changed(self):
        """
        Called when the edit button is clicked.
        Notifies the parent page to update the view accordingly.
        """
        self.parent().is_editable = not self.parent().is_editable
        self.parent().sync()


class ViewOnlyHeader(QtWidgets.QWidget):
    """
    A read-only header widget that displays impact category information.
    """

    def __init__(self, parent: ImpactCategoryHeader):
        """
        Initializes the view-only header.

        Args:
            parent (ImpactCategoryHeader): The parent header widget.
        """
        super().__init__(parent)
        self.grid = QtWidgets.QGridLayout()
        self.grid.setContentsMargins(0, 5, 0, 5)
        self.grid.setSpacing(10)
        self.grid.setColumnStretch(0, 0)  # Column 0 doesn't stretch
        self.grid.setColumnStretch(1, 1)  # Column 1 takes remaining space
        self.grid.setAlignment(QtCore.Qt.AlignmentFlag.AlignTop)
        self.setLayout(self.grid)
        
        # Create widgets
        self.name_label = QtWidgets.QLabel(self)
        self.unit_label = QtWidgets.QLabel(self)
        self.edit_button = QtWidgets.QPushButton("Edit Impact Category", self)
        self.edit_button.clicked.connect(parent.on_editable_changed)
        
        # Layout widgets
        self.grid.addWidget(widgets.ABLabel.demiBold("Name:", self), 0, 0)
        self.grid.addWidget(self.name_label, 0, 1)
        self.grid.addWidget(self.edit_button, 0, 1, QtCore.Qt.AlignmentFlag.AlignRight)
        self.grid.addWidget(widgets.ABLabel.demiBold("Unit:", self), 1, 0)
        self.grid.addWidget(self.unit_label, 1, 1)

    def sync(self):
        """
        Updates the displayed information from the current impact category.
        """
        impact_category = self.parent().impact_category
        self.name_label.setText(" | ".join(impact_category.name))
        self.unit_label.setText(impact_category.metadata.get("unit", "Undefined"))


class EditableHeader(QtWidgets.QWidget):
    """
    An editable header widget that allows modifying impact category information.
    """

    def __init__(self, parent: ImpactCategoryHeader):
        """
        Initializes the editable header.

        Args:
            parent (ImpactCategoryHeader): The parent header widget.
        """
        super().__init__(parent)
        self.grid = QtWidgets.QGridLayout()
        self.grid.setContentsMargins(0, 5, 0, 5)
        self.grid.setSpacing(10)
        self.grid.setColumnStretch(0, 0)  # Column 0 doesn't stretch
        self.grid.setColumnStretch(1, 1)  # Column 1 takes remaining space
        self.grid.setAlignment(QtCore.Qt.AlignmentFlag.AlignTop)
        self.setLayout(self.grid)
        
        # Create widgets
        self.name_label = QtWidgets.QLabel(self)
        self.name_label.linkActivated.connect(self._rename_method)
        self.unit_edit = ImpactCategoryUnit(self)
        self.done_button = QtWidgets.QPushButton("Done editing", self)
        self.done_button.clicked.connect(parent.on_editable_changed)
        
        # Layout widgets
        self.grid.addWidget(widgets.ABLabel.demiBold("Name:", self), 0, 0)
        self.grid.addWidget(self.name_label, 0, 1)
        self.grid.addWidget(self.done_button, 0, 1, QtCore.Qt.AlignmentFlag.AlignRight)
        self.grid.addWidget(widgets.ABLabel.demiBold("Unit:", self), 1, 0)
        self.grid.addWidget(self.unit_edit, 1, 1)

    def sync(self):
        """
        Updates the displayed information from the current impact category.
        """
        impact_category = self.parent().impact_category
        self.name_label.setText(f"<a href='/'>{' | '.join(impact_category.name)}</a>")
        self.unit_edit.setText(impact_category.metadata.get("unit", "Undefined"))

    def _rename_method(self):
        """
        Triggers the method rename action.
        """
        actions.MethodRename.run(self.parent().impact_category.name)


class ImpactCategoryUnit(QtWidgets.QLineEdit):
    """
    A line edit widget for editing the impact category unit.
    """

    def __init__(self, parent: EditableHeader):
        """
        Initializes the unit edit widget.

        Args:
            parent (EditableHeader): The parent editable header widget.
        """
        super().__init__(parent)
        self.editingFinished.connect(self.change_unit)

    def change_unit(self):
        """
        Updates the impact category unit when editing is finished.
        """
        impact_category = self.parent().parent().impact_category
        current_unit = impact_category.metadata.get("unit", "Undefined")
        
        if self.text() == current_unit:
            return
        
        actions.MethodMetaModify.run(impact_category.name, "unit", self.text())


from typing import List

from qtpy import QtWidgets


class ABDatabaseSelectionDialog(QtWidgets.QDialog):
    """Dialog to select one or more databases for export."""

    def __init__(self, parent=None, databases=None, title="Select databases"):
        super().__init__(parent=parent)
        self.setWindowTitle(title)
        self.setModal(True)
        self.resize(400, 300)

        layout = QtWidgets.QVBoxLayout(self)

        self.db_list_widget = QtWidgets.QListWidget(self)
        self.db_list_widget.setSelectionMode(QtWidgets.QAbstractItemView.MultiSelection)
        for db_name in databases:
            item = QtWidgets.QListWidgetItem(db_name)
            self.db_list_widget.addItem(item)
        layout.addWidget(self.db_list_widget)

        button_box = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel,
            parent=self
        )
        button_box.accepted.connect(self.accept)
        button_box.rejected.connect(self.reject)
        layout.addWidget(button_box)

    def get_selected_databases(self) -> List[str]:
        """Return the list of selected database names."""
        selected_items = self.db_list_widget.selectedItems()
        return [item.text() for item in selected_items]



import sqlite3
import pandas as pd
import pickle
import sys

def load(fp: str, database_name: str, fields: list[str]):
    con = sqlite3.connect(fp)
    sql = f"SELECT data FROM activitydataset WHERE database = '{database_name}'"
    raw_df = pd.read_sql(sql, con)
    con.close()

    df = pd.DataFrame([pickle.loads(x) for x in raw_df["data"]])
    if df.empty:
        return df

    df["key"] = list(zip(df["database"], df["code"]))
    df.index = pd.MultiIndex.from_tuples(df["key"], names=["database", "code"])
    df = df.reindex(columns=fields)[fields]
    return df

if __name__ == '__main__':
    filepath = sys.argv[1]
    database_name = sys.argv[2]
    columns = sys.argv[3:]
    df = load(filepath, database_name, columns)

    sys.stdout.buffer.write(pickle.dumps(df))



# -*- coding: utf-8 -*-
from itertools import zip_longest
from typing import List

from qtpy import QtCore, QtGui, QtWidgets


class OrderedListInputDialog(QtWidgets.QDialog):
    """Mostly cobbled together from: https://stackoverflow.com/a/41310284
    and https://stackoverflow.com/q/26936585
    """

    def __init__(self, parent=None, flags=QtCore.Qt.Window):
        super().__init__(parent=parent, f=flags)
        self.setWindowTitle("Select and order items")

        form = QtWidgets.QFormLayout(self)
        self.list_view = QtWidgets.QListView(self)
        self.list_view.setDragDropMode(QtWidgets.QListView.InternalMove)
        form.addRow(self.list_view)
        model = QtGui.QStandardItemModel(self.list_view)
        self.list_view.setModel(model)
        buttons = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel
        )
        form.addRow(buttons)
        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)
        self.show()

    @staticmethod
    def add_items_value(items, value: bool = False) -> List[tuple]:
        """Helper method, takes a list of items and adds given bool value,
        returning a list of tuples.
        """
        return [(i, b) for i, b in zip_longest(items, [], fillvalue=value)]

    def set_items(self, items: List[tuple]):
        model = self.list_view.model()
        model.clear()
        for i, checked in items:
            item = QtGui.QStandardItem(i)
            item.setCheckable(True)
            if checked:
                item.setCheckState(QtCore.Qt.CheckState.Checked)
            else:
                item.setCheckState(QtCore.Qt.CheckState.Unchecked)
            model.appendRow(item)
        self.list_view.setModel(model)

    def items_selected(self) -> list:
        model = self.list_view.model()
        selected = []
        for item in [model.item(i) for i in range(model.rowCount())]:
            if item.checkState():
                selected.append(item.text())
        return selected


class ListDelegate(QtWidgets.QStyledItemDelegate):
    """For managing and validating entered string values
    https://stackoverflow.com/a/40275439
    """
    def displayText(self, value, locale):
        if not isinstance(value, (list, tuple)):
            return str(value)
        return " > ".join(value)

    def createEditor(self, parent, option, index):
        editor = QtWidgets.QWidget(parent)
        dialog = OrderedListInputDialog(editor, QtCore.Qt.Window)
        dialog.accepted.connect(lambda: self.commitData.emit(editor))
        return editor

    def setEditorData(self, editor: QtWidgets.QWidget, index: QtCore.QModelIndex):
        """Populate the editor with data if editing an existing field."""
        dialog = editor.findChild(OrderedListInputDialog)
        value = index.data(QtCore.Qt.DisplayRole)
        values = [] if not value else [i.lstrip() for i in value.split(",")]

        parent = self.parent()
        if getattr(parent, "table_name", None) == "activity_parameter":
            groups = parent.get_activity_groups(index, values)
            unchecked = dialog.add_items_value(groups)
            checked = dialog.add_items_value(values, True)
            dialog.set_items(checked + unchecked)

    def setModelData(
        self,
        editor: QtWidgets.QWidget,
        model: QtCore.QAbstractItemModel,
        index: QtCore.QModelIndex,
    ):
        """Take the editor, read the given value and set it in the model"""
        dialog = editor.findChild(OrderedListInputDialog)
        value = dialog.items_selected()
        model.setData(index, value, QtCore.Qt.EditRole)



import os
from logging import getLogger

from qtpy import QtWidgets

from activity_browser import application
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.ui import icons, widgets
from activity_browser.bwutils.importers import ABPackage
from activity_browser.ui.core import threading

log = getLogger(__name__)


class DatabaseImporterBW2Package(ABAction):
    """ABAction to open the DatabaseImportWizard"""

    icon = icons.qicons.import_db
    text = "Import database from .bw2package"
    tool_tip = "Import database from .bw2package"

    @classmethod
    @exception_dialogs
    def run(cls):
        # get the path from the user
        path, _ = QtWidgets.QFileDialog.getOpenFileName(
            parent=application.main_window,
            caption='Choose .bw2package to import',
            filter='Brightway2 Database Package (*.bw2package);; All files (*.*)'
        )
        if not path:
            return

        # a bit of pathname magic to get a suggested database name
        context = {
            "path": path,
            "database_name": os.path.basename(path).split('.bw2package')[0]
        }

        # show the import setup dialog
        import_dialog = ImportSetup(parent=application.main_window, title="Import Database", context=context)
        import_dialog.exec_()


class ImportSetup(widgets.ABWizard):
    class DatabaseName(widgets.ABWizardPage):
        title = "Database Name"
        subtitle = "Enter the name of the database you wish to create"

        def __init__(self, parent=None):
            super().__init__(parent)
            self.db_name_edit = widgets.DatabaseNameEdit(
                label="Set database name:",
                database_preset="",
            )
            self.db_name_edit.textChanged.connect(self.completeChanged)

            layout = QtWidgets.QVBoxLayout()
            layout.addWidget(self.db_name_edit)
            self.setLayout(layout)

        def isComplete(self):
            return bool(self.db_name_edit.text())

        def initializePage(self, context: dict):
            self.db_name_edit.setText(context["database_name"])

        def finalize(self, context: dict):
            context["database_name"] = self.db_name_edit.text()

        def nextPage(self):
            return ImportSetup.InstallPage

    class InstallPage(widgets.ABThreadedWizardPage):
        """Wizard page to install the selected bw2package"""
        title = "Importing Database"
        subtitle = "Importing database from .bw2package file"

        class Thread(threading.ABThread):
            """Thread to handle the install process"""
            def run_safely(self, path: str, db_name: str):
                """Download the ecoinvent release"""
                ABPackage.import_file(path, rename=db_name)

        def initializePage(self, context: dict):
            """Start the download thread"""
            self.thread.start(context["path"], context["database_name"])

    pages = [DatabaseName, InstallPage]




from time import time
from logging import getLogger
from typing import Literal

import pandas as pd

from qtpy.QtCore import Qt, QObject, Signal, SignalInstance, QTimer

from .fields import all, all_types


log = getLogger(__name__)


class MetaDataStore(QObject):
    synced: SignalInstance = Signal(set, set, set)  # added, updated, deleted

    def __init__(self, parent=None):
        from activity_browser import application
        from .loader import MDSLoader
        from .updater import MDSUpdater

        super().__init__(parent)

        self._dataframe = pd.DataFrame()

        self._added: set[tuple[str, str]] = set()
        self._updated: set[tuple[str, str]] = set()
        self._deleted: set[tuple[str, str]] = set()

        self.loader = MDSLoader(self)
        self.updater = MDSUpdater(self)
        self.flusher: QTimer | None = None
        
        self.moveToThread(application.thread())

    @property
    def dataframe(self) -> pd.DataFrame:
        return self._dataframe

    @dataframe.setter
    def dataframe(self, df: pd.DataFrame) -> None:
        # Ensure all expected columns are present, in the correct order, and with the correct types
        df = df.reindex(columns=all)[all].astype(all_types)

        # Set the internal dataframe
        self._dataframe = df

    @property
    def databases(self):
        return set(self.dataframe.get("database", []))

    def register_mutation(self, key: tuple[str, str], action: Literal["add", "update", "delete"]):
        if action == "add":
            self._added.add(key)
            self._updated.discard(key)
            self._deleted.discard(key)

        elif action == "update":
            if key not in self._added:
                self._updated.add(key)

        elif action == "delete":
            if key in self._added:
                self._added.discard(key)
            else:
                self._deleted.add(key)
            self._updated.discard(key)
        else:
            raise ValueError(f"Unknown action: {action}")

        if not self.flusher:
            self.flusher = QTimer(self, interval=100)
            self.flusher.timeout.connect(self.flush_mutations)
            self.flusher.start()

    def flush_mutations(self):
        if not (self._added or self._updated or self._deleted):
            return

        t = time()
        self.synced.emit(self._added, self._updated, self._deleted)

        self._added.clear(), self._updated.clear(), self._deleted.clear()

        log.debug(f"Metadatastore sync signal completed in {time() - t:.2f} seconds")

    def match(self, **kwargs: dict[str, str]) -> pd.DataFrame:
        """Return a slice of the dataframe matching the criteria.
        """
        df = self.dataframe.query(
            " and ".join(
                [
                    f"`{key}` == '{value}'" if not pd.isna(value) else f"`{key}`.isnull()"
                    for key, value in kwargs.items()
                ])
        )

        return df

    def get_metadata(self, keys: list, columns: list) -> pd.DataFrame:
        """Return a slice of the dataframe matching row and column identifiers.

        NOTE: https://pandas.pydata.org/pandas-docs/stable/user_guide/indexing.html#deprecate-loc-reindex-listlike
        From pandas version 1.0 and onwards, attempting to select a column
        with all NaN values will fail with a KeyError.
        """
        df = self.dataframe.loc[pd.IndexSlice[keys], :]
        return df.reindex(columns, axis="columns")

    def get_database_metadata(self, db_name: str, columns: list = None) -> pd.DataFrame:
        if db_name not in self.databases:
            return pd.DataFrame(columns=all)
        return self.dataframe.loc[[db_name], columns or all]

AB_metadata = MetaDataStore()



from qtpy import QtWidgets, QtCore

import pandas as pd
import bw2data as bd
import bw_functional as bf

from activity_browser import actions
from activity_browser.bwutils import refresh_node, database_is_locked
from activity_browser.ui import widgets, delegates


class DataTab(QtWidgets.QWidget):
    """
    A widget that displays the data structure of a specific activity.

    Attributes:
        activity (tuple | int | bd.Node): The activity to display data for.
        data_view (DataView): The view displaying the data.
        data_model (DataModel): The model containing the data.
    """
    def __init__(self, activity: tuple | int | bd.Node, parent=None):
        """
        Initializes the DataTab widget.

        Args:
            activity (tuple | int | bd.Node): The activity to display data for.
            parent (QtWidgets.QWidget, optional): The parent widget. Defaults to None.
        """
        super().__init__(parent)

        self.activity = refresh_node(activity)

        # Data TreeView
        self.data_view = DataView(self)
        self.data_model = DataModel(self)
        self.data_view.setModel(self.data_model)

        self.data_model.setDataFrame(self.build_df())
        self.data_model.group(2)
        self.data_view.setColumnHidden(2, True)
        self.data_view.expandAll()

        self.build_layout()

    def build_layout(self):
        """
        Builds the layout of the widget.
        """
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.data_view)
        self.setLayout(layout)

    def sync(self) -> None:
        """
        Synchronizes the widget with the current state of the activity.
        """
        self.activity = refresh_node(self.activity)
        self.data_model.setDataFrame(self.build_df())

    def build_df(self) -> pd.DataFrame:
        """
        Builds a DataFrame from the activity data.

        Returns:
            pd.DataFrame: The DataFrame containing the activity data.
        """
        df = pd.Series(self.activity.as_dict()).to_frame()
        df["name"] = f"{self.activity['name']} {df.get('product', '')} ({self.activity['id']})"
        df["_activity_id"] = self.activity.id
        df["_activity_db"] = self.activity["database"]

        if isinstance(self.activity, bf.Process):
            for product in self.activity.products():
                fn_df = pd.DataFrame.from_dict(product.as_dict(), orient="index")
                fn_df["name"] = f"{product['name']}: {product.get('product', '')} ({product['id']})"
                fn_df["_activity_id"] = product.id
                fn_df["_activity_db"] = product["database"]
                df = pd.concat([df, fn_df])

        df = df.reset_index()
        df = df.rename({"index": "field", 0: "value"}, axis=1)
        df = df.sort_values(["name", "field"], ignore_index=True)

        cols = ["field", "value", "name", "_activity_id", "_activity_db"]
        return df[cols]


class DataView(widgets.ABTreeView):
    """
    A view that displays the data in a tree structure.

    Attributes:
        defaultColumnDelegates (dict): The default column delegates for the view.
    """
    defaultColumnDelegates = {
        "key": delegates.StringDelegate,
        "value": delegates.NewFormulaDelegate,
    }


class DataItem(widgets.ABDataItem):
    """
    An item representing a data entry in the tree view.
    """
    def flags(self, col: int, key: str):
        """
        Returns the item flags for the given column and key.

        Args:
            col (int): The column index.
            key (str): The key for which to return the flags.

        Returns:
            QtCore.Qt.ItemFlags: The item flags.
        """
        flags = super().flags(col, key)

        if key == "value" and not database_is_locked(self["_activity_db"]):
            return flags | QtCore.Qt.ItemFlag.ItemIsEditable
        return flags

    def displayData(self, col: int, key: str):
        """
        Returns the display data for the given column and key.

        Args:
            col (int): The column index.
            key (str): The key for which to return the display data.

        Returns:
            str: The display data.
        """
        if key == "value":
            data = self[key]
            if isinstance(data, str):
                return f"'{data}'"
            return str(data)

        return super().displayData(col, key)

    def setData(self, col: int, key: str, value) -> bool:
        """
        Sets the data for the given column and key.

        Args:
            col (int): The column index.
            key (str): The key for which to set the data.
            value: The value to set.

        Returns:
            bool: True if the data was set successfully, False otherwise.
        """
        if key in ["value"]:
            value = eval(value)
            actions.ActivityModify.run(self["_activity_id"], self["field"], value)

        return False


class DataModel(widgets.ABItemModel):
    """
    A model representing the data for the activity.

    Attributes:
        dataItemClass (type): The class of the data items.
    """
    dataItemClass = DataItem



from bw2io.migrations import *

from pyprind import ProgBar

def ab_create_core_migrations():
    """Activity Browser version of bw2io.migrations.create_core_migrations that employs a progress slot"""
    bar = ProgBar(12, title="Creating migrations")

    bar.title = "Creating migration: biosphere-2-3-categories"
    bar.update(0)
    Migration("biosphere-2-3-categories").write(
        get_biosphere_2_3_category_migration_data(),
        "Change biosphere category and subcategory labels to ecoinvent version 3",
    )

    bar.title = "Creating migration: biosphere-2-3-names"
    bar.update()
    Migration("biosphere-2-3-names").write(
        get_biosphere_2_3_name_migration_data(),
        "Change biosphere flow names to ecoinvent version 3",
    )

    bar.title = "Creating migration: simapro-ecoinvent-3.1"
    bar.update()
    Migration("simapro-ecoinvent-3.1").write(
        get_simapro_ecoinvent_3_migration_data("3.1"),
        "Change SimaPro names from ecoinvent 3.1 to ecoinvent names",
    )

    bar.title = "Creating migration: simapro-ecoinvent-3.2"
    bar.update()
    Migration("simapro-ecoinvent-3.2").write(
        get_simapro_ecoinvent_3_migration_data("3.2"),
        "Change SimaPro names from ecoinvent 3.2 to ecoinvent names",
    )

    bar.title = "Creating migration: simapro-ecoinvent-3.3"
    bar.update()
    Migration("simapro-ecoinvent-3.3").write(
        get_simapro_ecoinvent_3_migration_data("3.3"),
        "Change SimaPro names from ecoinvent 3.3 to ecoinvent names",
    )

    bar.title = "Creating migration: simapro-ecoinvent-3.4"
    bar.update()
    Migration("simapro-ecoinvent-3.4").write(
        get_simapro_ecoinvent_3_migration_data("3.4"),
        "Change SimaPro names from ecoinvent 3.4 to ecoinvent names",
    )

    bar.title = "Creating migration: simapro-water"
    bar.update()
    Migration("simapro-water").write(
        get_simapro_water_migration_data(),
        "Change SimaPro water flows to more standard names",
    )

    bar.title = "Creating migration: us-lci"
    bar.update()
    Migration("us-lci").write(
        get_us_lci_migration_data(), "Fix names in US LCI database"
    )

    bar.title = "Creating migration: default-units"
    bar.update()
    Migration("default-units").write(
        get_default_units_migration_data(), "Convert to default units"
    )

    bar.title = "Creating migration: unusual-units"
    bar.update()
    Migration("unusual-units").write(
        get_unusual_units_migration_data(), "Convert non-Ecoinvent units"
    )

    bar.title = "Creating migration: exiobase-biosphere"
    bar.update()
    Migration("exiobase-biosphere").write(
        get_exiobase_biosphere_migration_data(),
        "Change biosphere flow names to ecoinvent version 3",
    )

    bar.title = "Creating migration: fix-ecoinvent-flows-pre-35"
    bar.update()
    Migration("fix-ecoinvent-flows-pre-35").write(
        get_ecoinvent_pre35_migration_data(),
        "Update new biosphere UUIDs in Consequential 3.4",
    )
    bar.update()



from qtpy import QtWidgets

from activity_browser import application
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.ui import icons
from activity_browser.mod.bw2io.migrations import ab_create_core_migrations
from activity_browser.ui.core import threading


class MigrationsInstall(ABAction):
    """
    ABAction to install the default migrations from bw2io
    """

    icon = icons.qicons.import_db
    text = "Install default migrations"

    @staticmethod
    @exception_dialogs
    def run():
        def update_dialog_slot(progress: int, label: str):
            dialog.setValue(progress)
            dialog.setLabelText(label)


        dialog = QtWidgets.QProgressDialog(application.main_window)
        dialog.setWindowTitle("Installing migrations")
        dialog.setMaximum(100)
        dialog.setCancelButton(None)

        thread = MigrationsInstallThread(application)

        thread.status.connect(update_dialog_slot)

        thread.start()
        dialog.exec_()


class MigrationsInstallThread(threading.ABThread):
    def run_safely(self):
        ab_create_core_migrations()



from qtpy import QtCore, QtWidgets

from activity_browser import application
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.bwutils.strategies import relink_exchanges_existing_db
from activity_browser.mod import bw2data as bd
from activity_browser.ui.icons import qicons


class DatabaseRelink(ABAction):
    """
    ABAction to relink the dependencies of a database.
    """

    icon = qicons.edit
    text = "Relink the database"
    tool_tip = "Relink the dependencies of this database"

    @staticmethod
    @exception_dialogs
    def run(db_name: str):
        db_name = db_name
        # get brightway database object
        db = bd.Database(db_name)

        # find the dependencies of the database and construct a list of suitable candidates
        depends = db.find_dependents()
        options = [(depend, list(bd.databases)) for depend in depends]

        # construct a dialog in which the user chan choose which depending database to connect to which candidate
        dialog = DatabaseLinkingDialog.relink_sqlite(
            db_name, options, application.main_window
        )

        # return if the user cancels
        if dialog.exec_() != DatabaseLinkingDialog.Accepted:
            return

        # else, start the relinking
        QtWidgets.QApplication.setOverrideCursor(QtCore.Qt.WaitCursor)
        relinking_results = dict()

        # relink using relink_exchanges_existing_db strategy
        for old, new in dialog.relink.items():
            other = bd.Database(new)
            failed, succeeded, examples = relink_exchanges_existing_db(db, old, other)
            relinking_results[f"{old} --> {other.name}"] = (failed, succeeded)

        QtWidgets.QApplication.restoreOverrideCursor()

        # if any failed, present user with results dialog
        if failed > 0:
            relinking_dialog = DatabaseLinkingResultsDialog.present_relinking_results(
                application.main_window, relinking_results, examples
            )
            relinking_dialog.exec_()
            relinking_dialog.open_activity()


class DatabaseLinkingDialog(QtWidgets.QDialog):
    """Display all of the possible links in a single dialog for the user.

    Allow users to select alternate database links."""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Database linking")

        self.db_label = QtWidgets.QLabel()
        self.label_choices = []
        self.grid_box = QtWidgets.QGroupBox("Database links:")
        self.grid = QtWidgets.QGridLayout()
        self.grid_box.setLayout(self.grid)

        self.buttons = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel,
        )
        self.buttons.accepted.connect(self.accept)
        self.buttons.rejected.connect(self.reject)

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.db_label)
        layout.addWidget(self.grid_box)
        layout.addWidget(self.buttons)
        self.setLayout(layout)

    @property
    def relink(self) -> dict:
        """Returns a dictionary of str -> str key/values, showing which keys
        should be linked to which values.

        Only returns key/value pairs if they differ.
        """
        return {
            label.text(): combo.currentText()
            for label, combo in self.label_choices
            if label.text() != combo.currentText()
        }

    @property
    def links(self) -> dict:
        """Returns a dictionary of str -> str key/values, showing which keys
        should be linked to which values.
        """
        return {
            label.text(): combo.currentText() for label, combo in self.label_choices
        }

    @classmethod
    def construct_dialog(
        cls,
        label: str,
        options: list,
        parent: QtWidgets.QWidget = None,
    ) -> "DatabaseLinkingDialog":
        obj = cls(parent)
        obj.db_label.setText(label)
        # Start at 1 because row 0 is taken up by the db_label
        for i, item in enumerate(options):
            label = QtWidgets.QLabel(item[0])
            combo = QtWidgets.QComboBox()
            combo.addItems(item[1])
            combo.setCurrentText(item[0])
            obj.label_choices.append((label, combo))
            obj.grid.addWidget(label, i, 0, 1, 2)
            obj.grid.addWidget(combo, i, 2, 1, 2)
        obj.updateGeometry()
        return obj

    @classmethod
    def relink_sqlite(
        cls, db: str, options: list, parent=None
    ) -> "DatabaseLinkingDialog":
        label = "Relinking exchanges from database '{}'.".format(db)
        return cls.construct_dialog(label, options, parent)

    @classmethod
    def relink_bw2package(
        cls, options: list, parent=None
    ) -> "DatabaseLinkingDialog":
        label = (
            "Some database(s) could not be found in the current project,"
            " attempt to relink the exchanges to a different database?"
        )
        return cls.construct_dialog(label, options, parent)

    @classmethod
    def relink_excel(
        cls, options: list, parent=None
    ) -> "DatabaseLinkingDialog":
        label = "Customize database links for exchanges in the imported database."
        return cls.construct_dialog(label, options, parent)


class DatabaseLinkingResultsDialog(QtWidgets.QDialog):
    """To be used when relinking a database, this dialog will pop up if
    some of the exchanges in the database fail to be linked to the new
    database.
    Up to five of the unlinked activities are printed on the screen,

    """

    def __init__(self, parent=None):
        super().__init__(parent)

        self.setWindowTitle("Relinking database results")

        button = QtWidgets.QDialogButtonBox.Ok
        self.buttonBox = QtWidgets.QDialogButtonBox(button)
        self.buttonBox.accepted.connect(self.accept)
        self.databases_relinked = QtWidgets.QVBoxLayout()

        self.activityToOpen = set()

        self.exchangesUnlinked = QtWidgets.QVBoxLayout()

        self.layout = QtWidgets.QVBoxLayout()
        self.layout.addLayout(self.databases_relinked)
        self.layout.addLayout(self.exchangesUnlinked)
        self.layout.addWidget(self.buttonBox)
        self.setLayout(self.layout)

    @classmethod
    def construct_results_dialog(
        cls,
        parent: QtWidgets.QWidget = None,
        link_results: dict = None,
        unlinked_exchanges: dict = None,
    ) -> "DatabaseLinkingResultsDialog":
        from activity_browser import actions

        obj = cls(parent)
        for k, results in link_results.items():
            obj.databases_relinked.addWidget(
                QtWidgets.QLabel(f"{k} = {results[1]} successfully linked")
            )
            obj.databases_relinked.addWidget(
                QtWidgets.QLabel(f"{k} = {results[0]} flows failed to link")
            )

        obj.exchangesUnlinked.addWidget(
            QtWidgets.QLabel("Up to 5 unlinked exchanges (click to open)")
        )
        for act, key in unlinked_exchanges.items():
            button = QtWidgets.QPushButton(act.as_dict()["name"])
            button.clicked.connect(
                lambda: actions.ActivityOpen.run([act.key])
            )
            obj.exchangesUnlinked.addWidget(button)
        obj.updateGeometry()

        return obj

    @classmethod
    def present_relinking_results(
        cls,
        parent: QtWidgets.QWidget = None,
        link_results: dict = None,
        unlinked_exchanges: dict = None,
    ) -> "DatabaseLinkingResultsDialog":
        return cls.construct_results_dialog(parent, link_results, unlinked_exchanges)

    def select_activity_to_open(self, actvty: tuple) -> None:
        if actvty in self.activityToOpen:
            self.activityToOpen.discard(actvty)
        self.activityToOpen.add(actvty)

    def open_activity(self):
        return self.activityToOpen




from uuid import uuid4

from qtpy.QtWidgets import QDialog
import bw2data as bd

from activity_browser import application, bwutils
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.ui.icons import qicons
from activity_browser.ui.widgets.new_node_dialog import NewNodeDialog

from .activity_open import ActivityOpen


class ActivityNewProcess(ABAction):
    """
    ABAction to create a new activity. Prompts the user to supply a name. Returns if no name is supplied or if the user
    cancels. Otherwise, instructs the ActivityController to create a new activity.
    """

    icon = qicons.add
    text = "New process"

    @staticmethod
    @exception_dialogs
    def run(database_name: str):
        # ask the user to provide a name for the new activity
        dialog = NewNodeDialog(application.main_window)
        # if the user cancels, return
        if dialog.exec_() != QDialog.Accepted:
            return
        name, ref_product, unit, location = dialog.get_new_process_data()
        # if no name is provided, return
        if not name:
            return
        if ref_product == "":
            ref_product = name

        database = bd.Database(database_name)
        legacy_backend = bwutils.database_is_legacy(database_name)

        # create process
        new_proc_data = {
            "name": name,
            "location": location,
            "type": "process" if not legacy_backend else "processwithreferenceproduct",
        }

        if legacy_backend:
            new_proc_data["reference product"] = ref_product
            new_proc_data["unit"] = unit

        new_process: bd.Node = database.new_activity(code=uuid4().hex, **new_proc_data)
        new_process.save()

        if legacy_backend:
            new_process.new_exchange(
                input=new_process.key,
                type="production",
                amount=1.0,
            ).save()

        if not legacy_backend:
            # create reference product
            new_ref_prod_data = {
                "product": ref_product,
                "unit": unit,
                "location": location,
                "type": "product",
            }
            prod = new_process.new_product(code=uuid4().hex, **new_ref_prod_data)
            prod.save()

        ActivityOpen.run([new_process.key])



from typing import List

from qtpy import QtCore, QtWidgets

import bw2data as bd
from bw2data.parameters import Group
from bw2data.backends.proxies import ExchangeDataset, Exchanges

from activity_browser import application, settings
from activity_browser.bwutils import AB_metadata
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.ui.icons import qicons


class DatabaseDelete(ABAction):
    """
    Deletes one or more databases from the project after user confirmation.

    This method performs the following steps:
    - Displays a confirmation dialog to the user with the database name(s) and total record count.
    - If the user confirms, deletes the database(s), their upstream exchanges, and associated parameters.
    - Removes the database(s) from the project settings.

    Args:
        db_names (List[str]): The name(s) of the database(s) to be deleted.

    Steps:
    - Set the cursor to a waiting state while gathering data for large databases.
    - Retrieve the record count for the specified database(s).
    - Construct a warning message with the database name(s) and record count.
    - Display a confirmation dialog to the user.
    - If the user cancels, exit the method.
    - Set the cursor to a waiting state while performing the deletion.
    - Delete upstream exchanges associated with the database(s).
    - Remove the database(s) from the Brightway2 project.
    - Delete database parameters.
    - Remove the database(s) from the project settings.
    - Restore the cursor to its default state.
    """

    icon = qicons.delete
    text = "Delete databases"
    tool_tip = "Delete database(s) from the project"

    @staticmethod
    @exception_dialogs
    def run(db_names: List[str]):
        # gathering data will take time for large databases
        QtWidgets.QApplication.setOverrideCursor(QtCore.Qt.WaitCursor)

        # get the total record count from all databases
        total_records = 0
        for db_name in db_names:
            n_records = AB_metadata.dataframe[AB_metadata.dataframe["database"] == db_name].shape[0]
            total_records += n_records

        # construct warning text
        if len(db_names) == 1:
            text = f"Are you sure you want to delete database <b>'{db_names[0]}'</b>?"
            if total_records:
                text += f" It contains {total_records} activities."
        else:
            text = f"Are you sure you want to delete {len(db_names)} databases?"
            if total_records:
                text += f" They contain {total_records} activities in total."

        # ask the user for confirmation
        QtWidgets.QApplication.restoreOverrideCursor()
        response = QtWidgets.QMessageBox.question(
            application.main_window, build_title(db_names), text
        )

        # return if the user cancels
        if response != QtWidgets.QMessageBox.Yes:
            return

        # deleting data will take time for large databases
        QtWidgets.QApplication.setOverrideCursor(QtCore.Qt.WaitCursor)

        for db_name in db_names:
            # delete upstream exchanges
            ExchangeDataset.delete().where(ExchangeDataset.input_database == db_name).execute()

            # instruct the DatabaseController to delete the database from the project.
            del bd.databases[db_name]

            # delete database parameters
            Group.delete().where(Group.name == db_name).execute()

            # remove database from project settings
            settings.project_settings.remove_db(db_name)

        QtWidgets.QApplication.restoreOverrideCursor()


def build_title(db_names: List[str]) -> str:
    """Build an appropriate title for the confirmation dialog."""
    if len(db_names) == 1:
        return "Delete database?"
    return "Delete databases?"



from qtpy import QtWidgets, QtGui
from qtpy.QtCore import Qt

import bw2data as bd
import pandas as pd

from activity_browser import actions, signals
from activity_browser.ui import widgets, icons, delegates
from activity_browser.bwutils import AB_metadata, is_node_biosphere

from .impact_category_header import ImpactCategoryHeader


class ImpactCategoryDetailsPage(QtWidgets.QWidget):
    def __init__(self, name: tuple, parent=None):
        super().__init__(parent)
        self.name = name
        self.impact_category = bd.Method(name)
        self.is_editable = False

        self.setObjectName(" | ".join(name))

        self.header = ImpactCategoryHeader(self)

        self.model = CharacterizationFactorsModel(self)
        self.view = CharacterizationFactorsView(self)
        self.view.setModel(self.model)
        self.view.setSortingEnabled(True)

        self.build_layout()
        self.connect_signals()
        self.sync()

        # resizing name and categories columns
        self.view.resizeColumnToContents(0)
        self.view.resizeColumnToContents(1)

    def connect_signals(self):
        signals.method.renamed.connect(self.on_method_renamed)
        signals.method.deleted.connect(self.on_method_deleted)
        signals.meta.methods_changed.connect(self.sync)

    def on_method_renamed(self, old_name, new_name):
        if self.name == old_name:
            self.name = new_name
            self.setObjectName(" | ".join(new_name))
            self.setWindowTitle(" | ".join(new_name))

    def on_method_deleted(self, method):
        if method.name == self.name:
            self.deleteLater()

    def sync(self):
        if self.name not in bd.methods:
            self.deleteLater()
            return

        self.impact_category = bd.Method(self.name)
        self.model.setDataFrame(self.build_df())
        self.header.sync()

    def build_layout(self):
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.header)
        layout.addWidget(widgets.ABHLine(self))
        layout.addWidget(self.view)
        self.setLayout(layout)

    def build_df(self):
        df = pd.DataFrame(self.impact_category.load(), columns=["id", "data"])
        df["amount"] = df["data"].apply(lambda x: x if isinstance(x, (float, int)) else x.get("amount"))
        df["uncertainty"] = df["data"].apply(lambda x: 0 if isinstance(x, (float, int)) else x.get("uncertainty type"))

        other = AB_metadata.dataframe[["id", "name", "categories", "database", "unit"]]

        df = df.merge(other, left_on="id", right_on="id").rename(columns={"id": "_id", "data": "_cf"})
        df["_impact_category_name"] = [self.name for i in range(len(df))]
        df["_editable"] = self.is_editable

        cols = ["name", "categories", "database", "amount", "unit", "uncertainty", "_id", "_impact_category_name", "_cf", "_editable"]
        return df[cols]


class CharacterizationFactorsView(widgets.ABTreeView):
    defaultColumnDelegates = {
        "amount": delegates.FloatDelegate,
        "categories": delegates.ListDelegate,
        "uncertainty": delegates.UncertaintyDelegate,
    }

    class ContextMenu(widgets.ABMenu):
        menuSetup = [
            lambda m: m.add(actions.CFRemove, m.impact_category_name, m.char_factors,
                            enable=bool(m.char_factors) and m.is_editable,
                            text="Remove characterization factor(s)"),
        ]

        @property
        def is_editable(self):
            return self.parent().parent().is_editable

        @property
        def impact_category_name(self):
            return self.parent().parent().name

        @property
        def char_factors(self):
            indexes = self.parent().selectedIndexes()
            return [(idx.internalPointer()["_id"], idx.internalPointer()["_cf"]) 
                    for idx in indexes if idx.isValid() and idx.column() == 0]

    def __init__(self, parent):
        super().__init__(parent)
        self.setAcceptDrops(True)
        self.setSortingEnabled(True)
        self.overlay = None

    def dragEnterEvent(self, event):
        """
        Handles the drag enter event.

        Args:
            event: The drag enter event.
        """
        if not self.parent().is_editable:
            event.ignore()
            return

        if event.mimeData().hasFormat("application/bw-nodekeylist"):
            self.overlay = widgets.ABDropOverlay(self)
            self.overlay.show()
            event.accept()
        else:
            event.ignore()
    
    def dragMoveEvent(self, event):
        """Handles the drag move event - required for proper drop indicator."""
        if not self.parent().is_editable:
            event.ignore()
            return

        if event.mimeData().hasFormat("application/bw-nodekeylist"):
            event.accept()
        else:
            event.ignore()

    def dragLeaveEvent(self, event):
        """
        Handles the drag leave event.

        Args:
            event: The drag leave event.
        """
        if not self.overlay is None:
            # Reset the palette on drag leave
            self.overlay.deleteLater()
            self.overlay = None

    def dropEvent(self, event):
        """
        Handles the drop event.

        Args:
            event: The drop event.
        """
        self.overlay.deleteLater()
        self.overlay = None

        keys: list = event.mimeData().retrievePickleData("application/bw-nodekeylist")

        # Filter to only biosphere flows
        biosphere_keys = [key for key in keys if is_node_biosphere(key)]

        if biosphere_keys:
            actions.CFNew.run(self.parent().name, biosphere_keys)


class ExchangesItem(widgets.ABDataItem):
    def flags(self, col: int, key: str):
        """
        Returns the item flags for the given column and key.

        Args:
            col (int): The column index.
            key (str): The key for which to return the flags.

        Returns:
            QtCore.Qt.ItemFlags: The item flags.
        """
        flags = super().flags(col, key)

        if key in ["amount", "uncertainty"] and self["_editable"]:
            return flags | Qt.ItemFlag.ItemIsEditable
        return flags

    def decorationData(self, col, key):
        """
        Provides decoration data for the item.

        Args:
            col: The column index.
            key: The key for which to provide decoration data.

        Returns:
            The decoration data for the item.
        """
        if key == "name":
            return icons.qicons.biosphere

    def fontData(self, col: int, key: str):
        """
        Returns the font data for the given column and key.

        Args:
            col (int): The column index.
            key (str): The key for which to return the font data.

        Returns:
            QtGui.QFont: The font data.
        """
        font = super().fontData(col, key)

        # set the font to bold if it's a production/functional exchange
        if key == "name":
            font.setWeight(QtGui.QFont.Weight.DemiBold)
        return font

    def setData(self, col: int, key: str, value) -> bool:
        """
        Sets the data for the given column and key.

        Args:
            col (int): The column index.
            key (str): The key for which to set the data.
            value: The value to set.

        Returns:
            bool: True if the data was set successfully, False otherwise.
        """
        if key not in ["amount"]:
            return False

        actions.CFAmountModify.run(self["_impact_category_name"], self["_id"], value)


class CharacterizationFactorsModel(widgets.ABItemModel):
    dataItemClass = ExchangesItem



from qtpy import QtCore, QtGui, QtWidgets
from activity_browser.ui.icons import qicons


class DragHandleDelegate(QtWidgets.QStyledItemDelegate):
    """Custom delegate that paints a drag handle icon on the left of each row.
    
    This delegate adds a visual affordance (grip icon) to indicate that rows
    can be reordered via drag-and-drop. The icon is painted in the left margin
    of each list item.
    """
    
    def paint(self, painter, option, index):
        """Paint the item with a drag handle icon on the left side.
        
        Parameters
        ----------
        painter : QtGui.QPainter
            The painter to use for rendering.
        option : QtWidgets.QStyleOptionViewItem
            Style options for the item.
        index : QtCore.QModelIndex
            The model index of the item to paint.
        """
        super().paint(painter, option, index)
        
        # Draw drag handle icon on the left
        icon_size = 16
        icon_margin = 4
        icon_rect = QtCore.QRect(
            option.rect.left() + icon_margin,
            option.rect.top() + (option.rect.height() - icon_size) // 2,
            icon_size,
            icon_size
        )
        
        # Use a grip icon if available, otherwise use a simple visual indicator
        if hasattr(qicons, 'drag_indicator'):
            qicons.drag_indicator.paint(painter, icon_rect)
        else:
            # Fallback: draw three horizontal lines as a grip indicator
            painter.save()
            painter.setPen(QtGui.QPen(option.palette.mid().color(), 2))
            y_center = icon_rect.center().y()
            x_left = icon_rect.left() + 2
            x_right = icon_rect.right() - 2
            for offset in [-4, 0, 4]:
                y = y_center + offset
                painter.drawLine(x_left, y, x_right, y)
            painter.restore()


class ABListEditDialog(QtWidgets.QDialog):
    """
    A dialog for editing a list or tuple of strings with drag-and-drop reordering.
        
    Parameters
    ----------
    data : iterable of str
        Initial values to populate the list.
    title : str, optional
        Window title for the dialog. Default is "Edit List/Tuple".
    parent : QtWidgets.QWidget, optional
        Parent widget for the dialog.
    
    Examples
    --------
    >>> dialog = ABListEditDialog(["item1", "item2"], title="Edit Items")
    >>> if dialog.exec_() == QtWidgets.QDialog.Accepted:
    ...     updated_items = dialog.get_data()
    """

    def __init__(self, data, title="Edit List/Tuple", parent=None):
        super().__init__(parent)
        self.setWindowTitle(title)
        self.resize(420, 320)

        layout = QtWidgets.QVBoxLayout(self)

        # List widget
        self.list = QtWidgets.QListWidget(self)
        self.list.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.list.setEditTriggers(
            QtWidgets.QAbstractItemView.DoubleClicked | QtWidgets.QAbstractItemView.EditKeyPressed
        )
        # Enable intuitive drag-and-drop reordering
        self.list.setDragDropMode(QtWidgets.QAbstractItemView.InternalMove)
        self.list.setDefaultDropAction(QtCore.Qt.MoveAction)
        self.list.setAlternatingRowColors(True)
        self.list.setStyleSheet(
            """
            QListWidget { alternate-background-color: palette(alternate-base); }
            QListWidget::item { padding: 6px 28px 6px 28px; }
            QListWidget::item:selected { background: palette(highlight); color: palette(highlighted-text); }
            """
        )
        # Set custom delegate to draw drag handles
        self.list.setItemDelegate(DragHandleDelegate(self.list))

        # OK/Cancel
        self.button_box = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel
        )

        # Assemble layout
        layout.addWidget(self.list)
        layout.addWidget(self.button_box)
        self.setLayout(layout)

        # Signals
        self.button_box.accepted.connect(self.accept)
        self.button_box.rejected.connect(self.reject)
        self.list.itemChanged.connect(self.on_item_changed)
        self.list.itemSelectionChanged.connect(self._on_selection_changed)
        # Reposition inline buttons on scroll/resize/content changes
        self.list.verticalScrollBar().valueChanged.connect(self._position_inline_buttons)
        self.list.horizontalScrollBar().valueChanged.connect(self._position_inline_buttons)
        self.list.viewport().installEventFilter(self)

        # Populate from provided data
        self.load_data(data)
        self._create_inline_buttons()
        self._position_inline_buttons()

    # ---------- Data ----------
    def load_data(self, data):
        """Load data into the list widget.
        
        Populates the list with the provided values. If no data is provided,
        adds a single empty row to guide the user.
        
        Parameters
        ----------
        data : iterable of str
            Values to populate the list with.
        """
        has_any = False
        for value in data:
            self._append_item(str(value))
            has_any = True
        if not has_any:
            # Provide a single empty row to guide the user
            self._append_item("")

    def get_data(self, as_tuple=False):
        """Retrieve the current list data, excluding empty rows.
        
        Parameters
        ----------
        as_tuple : bool, optional
            If True, return data as a tuple instead of a list. Default is False.
        
        Returns
        -------
        list or tuple of str
            Non-empty string values from the list, in current display order.
        """
        values = []
        for row in range(self.list.count()):
            item = self.list.item(row)
            if item:
                text = item.text().strip()
                if text:
                    values.append(text)
        return tuple(values) if as_tuple else values

    # ---------- Item helpers ----------
    def _append_item(self, text=""):
        """Create and append a new editable, draggable item to the list.
        
        Parameters
        ----------
        text : str, optional
            Initial text content for the item. Default is empty string.
        """
        item = QtWidgets.QListWidgetItem(text)
        # editable + enabled + selectable + draggable
        item.setFlags(item.flags() | QtCore.Qt.ItemIsEditable | QtCore.Qt.ItemIsDragEnabled)
        self.list.addItem(item)

    def add_item(self):
        """Add a new empty item and immediately start editing it.
        
        This method is connected to the inline add button. It creates a new row,
        selects it, and opens it for editing, then repositions the floating buttons.
        """
        self._append_item("")
        # Select and start editing the newly added row
        item = self.list.item(self.list.count() - 1)
        self.list.setCurrentItem(item)
        self.list.editItem(item)
        self._position_inline_buttons()

    def remove_selected(self):
        """Remove the currently selected item from the list.
        
        This method is connected to the inline remove button. After removal,
        it ensures at least one empty row remains and adjusts the selection
        to the next appropriate row.
        """
        row = self._current_row()
        if row is None:
            return
        self.list.takeItem(row)
        # keep at least one empty row for guidance
        if self.list.count() == 0:
            self._append_item("")
        # adjust selection
        new_row = min(row, self.list.count() - 1)
        if new_row >= 0:
            self.list.setCurrentRow(new_row)
        self._position_inline_buttons()

    # Drag-and-drop handles reordering; no explicit move buttons

    def on_item_changed(self, item: QtWidgets.QListWidgetItem):
        """Handle item text changes by normalizing whitespace.
        
        Connected to the list widget's itemChanged signal. Collapses multiple
        consecutive spaces into a single space to maintain clean data.
        
        Parameters
        ----------
        item : QtWidgets.QListWidgetItem
            The item that was changed.
        """
        # No placeholder logic. Just normalize whitespace.
        if item is None:
            return
        text = item.text()
        if text is None:
            return
        # Collapse accidental multiple spaces
        norm = " ".join(text.split())
        if norm != text:
            # block signals to avoid recursion
            self.list.blockSignals(True)
            item.setText(norm)
            self.list.blockSignals(False)

    # ---------- UI helpers ----------
    def _current_row(self):
        """Get the index of the currently selected row.
        
        Returns
        -------
        int or None
            Row index (0-based) of the selected item, or None if nothing is selected.
        """
        indexes = self.list.selectedIndexes()
        if not indexes:
            return None
        return indexes[0].row()

    # ---------- Inline buttons ----------
    def _create_inline_buttons(self):
        """Create floating icon buttons for add and remove operations.
        
        Creates two QToolButton instances:
        - Add button: positioned at bottom-right corner of the list viewport
        - Remove button: positioned inline with the currently selected row
        
        Both buttons use absolute positioning and are parented to the viewport
        to float over the list content.
        """
        # Add button at bottom-right
        self.inline_add_btn = QtWidgets.QToolButton(self.list.viewport())
        self.inline_add_btn.setIcon(qicons.add)
        self.inline_add_btn.setAutoRaise(True)
        self.inline_add_btn.setToolTip("Add row")
        self.inline_add_btn.clicked.connect(self.add_item)
        
        # Remove button aligned with selected row
        self.inline_remove_btn = QtWidgets.QToolButton(self.list.viewport())
        self.inline_remove_btn.setIcon(qicons.delete)
        self.inline_remove_btn.setAutoRaise(True)
        self.inline_remove_btn.setToolTip("Remove selected row")
        self.inline_remove_btn.clicked.connect(self.remove_selected)
        self.inline_remove_btn.hide()

    def _on_selection_changed(self):
        """Handle selection changes by repositioning inline buttons.
        
        Connected to the list widget's itemSelectionChanged signal. Ensures
        the remove button follows the selected row.
        """
        self._position_inline_buttons()

    def eventFilter(self, obj, event):
        """Monitor viewport events to reposition floating buttons when needed.
        
        Watches for resize, update, and paint events on the list viewport,
        deferring button repositioning until after layout updates complete.
        
        Parameters
        ----------
        obj : QtCore.QObject
            The object being monitored (should be self.list.viewport()).
        event : QtCore.QEvent
            The event that occurred.
        
        Returns
        -------
        bool
            Result from the parent event filter.
        """
        if obj is self.list.viewport():
            if event.type() in (QtCore.QEvent.Resize, QtCore.QEvent.UpdateRequest, QtCore.QEvent.Paint):
                # Defer reposition slightly to after layout updates
                QtCore.QTimer.singleShot(0, self._position_inline_buttons)
        return super().eventFilter(obj, event)

    def _position_inline_buttons(self):
        """Calculate and apply absolute positions for floating buttons.
        
        Positions the add button at the bottom-right corner of the viewport,
        and the remove button inline with the currently selected row (if any).
        The remove button is only shown if it would be visible within the viewport.
        
        This method is called on:
        - Selection changes
        - Scroll events
        - Viewport resize/paint/update events
        - After adding or removing items
        """
        if not hasattr(self, "inline_add_btn") or not hasattr(self, "inline_remove_btn"):
            return
        
        # Position add button at bottom-right corner
        viewport_rect = self.list.viewport().rect()
        add_w = self.inline_add_btn.sizeHint().width()
        add_h = self.inline_add_btn.sizeHint().height()
        add_x = viewport_rect.right() - add_w - 6
        add_y = viewport_rect.bottom() - add_h - 6
        self.inline_add_btn.move(add_x, add_y)
        self.inline_add_btn.show()
        
        # Position remove button aligned with selected row
        row = self._current_row()
        if row is None:
            self.inline_remove_btn.hide()
            return
        item = self.list.item(row)
        if item is None:
            self.inline_remove_btn.hide()
            return
        rect = self.list.visualItemRect(item)
        if not rect.isValid() or rect.height() <= 0:
            self.inline_remove_btn.hide()
            return
        # Position inside the item's rect at right side with small margin
        btn_w = self.inline_remove_btn.sizeHint().width()
        btn_h = self.inline_remove_btn.sizeHint().height()
        x = rect.right() - btn_w - 6
        y = rect.top() + (rect.height() - btn_h) // 2
        self.inline_remove_btn.move(x, y)
        # Only show if fully or partially visible within viewport
        if viewport_rect.intersects(QtCore.QRect(x, y, btn_w, btn_h)):
            self.inline_remove_btn.show()
        else:
            self.inline_remove_btn.hide()





from logging import getLogger

import pandas as pd
import bw2data as bd

from qtpy import QtCore, QtWidgets

from activity_browser import application, bwutils
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.ui.icons import qicons

log = getLogger(__name__)


class CSCalculate(ABAction):
    """
    ABAction to calculate a calculation setup. First asks the user for confirmation and returns if cancelled. Otherwise,
    passes the csname to the CalculationSetupController for calculation. Finally, displays confirmation that it succeeded.
    """

    icon = qicons.calculate
    text = "Calculate"

    @staticmethod
    @exception_dialogs
    def run(cs_name: str, scenario_data: pd.DataFrame = None):
        from activity_browser.layouts import pages

        # Check if the calculation setup is complete
        if cs_name not in bd.calculation_setups:
            raise Exception(f"Calculation setup '{cs_name}' not found.")
        cs = bd.calculation_setups[cs_name]
        if not cs.get("inv"):
            raise Exception(f"Calculation setup '{cs_name}' has no functional units.")
        if not cs.get("ia"):
            raise Exception(f"Calculation setup '{cs_name}' has no impact assessment methods.")

        dialog = CalculationDialog(cs_name, application.main_window)
        dialog.show()
        application.thread().eventDispatcher().processEvents(QtCore.QEventLoop.ProcessEventsFlag.AllEvents)

        try:
            if scenario_data is None:
                mlca = bwutils.MLCA(cs_name)
                contributions = bwutils.Contributions(mlca)
            else:
                mlca = bwutils.SuperstructureMLCA(cs_name, scenario_data)
                contributions = bwutils.SuperstructureContributions(mlca)

            mlca.calculate()
            mc = bwutils.MonteCarloLCA(cs_name)

            page = pages.LCAResultsPage(cs_name, mlca, contributions, mc)
            central = application.main_window.centralWidget()
        except:
            dialog.close()
            raise

        dialog.close()
        central.addToGroup("LCA Results", page)


class CalculationDialog(QtWidgets.QDialog):
    def __init__(self, cs_name: str, parent=None):
        super().__init__(parent, QtCore.Qt.WindowTitleHint)
        self.setWindowTitle(f"Running Calculations")
        self.setModal(True)

        self.label = QtWidgets.QLabel(f"Running calculations for setup: <b>{cs_name}</b>", self)

        layout = QtWidgets.QVBoxLayout(self)
        layout.addWidget(self.label)
        self.setLayout(layout)



import os
from zipfile import ZipFile
from logging import getLogger

from bw2io.importers import Ecospold2BiosphereImporter
from bw2io.importers.ecospold2_biosphere import EMISSIONS_CATEGORIES
from lxml import objectify

from activity_browser.mod import bw2data as bd

from ...info import __ei_versions__
from ...utils import sort_semantic_versions

log = getLogger(__name__)


def create_default_biosphere3(version) -> None:
    """Reimplementation of bw.create_default_biosphere3 to allow import from older biosphere versions."""
    # format version number to only Major/Minor
    version = version[:3]

    if version == sort_semantic_versions(__ei_versions__)[0][:3]:
        log.debug(f"Installing biosphere version >{version}<")
        # most recent version
        eb = Ecospold2BiosphereImporter()
    else:
        log.debug(f"Installing legacy biosphere version >{version}<")
        # not most recent version, import legacy biosphere from AB
        eb = ABEcospold2BiosphereImporter(version=version)
    eb.apply_strategies()
    eb.write_database()


class ABEcospold2BiosphereImporter(Ecospold2BiosphereImporter):
    """Reimplementation of bw2io.importers Ecospold2BiosphereImporter to import legacy biosphere from AB data"""

    def extract(self, version, filepath=None):
        def extract_flow_data(o):
            ds = {
                "categories": (
                    o.compartment.compartment.text,
                    o.compartment.subcompartment.text,
                ),
                "code": o.get("id"),
                "CAS number": o.get("casNumber"),
                "name": o.name.text,
                "database": self.db_name,
                "exchanges": [],
                "unit": o.unitName.text,
            }
            ds["type"] = EMISSIONS_CATEGORIES.get(
                ds["categories"][0], ds["categories"][0]
            )
            return ds

        lci_dirpath = os.path.join(os.path.dirname(__file__), "legacy_biosphere")

        # find the most recent legacy biosphere that is equal to or older than chosen version
        for ei_version in sort_semantic_versions(__ei_versions__):
            use_version = ei_version
            fp = os.path.join(
                lci_dirpath, f"ecoinvent elementary flows {use_version}.xml.zip"
            )
            if sort_semantic_versions([version, ei_version])[
                0
            ] == version and os.path.isfile(fp):
                # this version is equal/lower and available
                break

        # extract the xml from the zip
        with ZipFile(fp) as zipped_file:
            with zipped_file.open(
                f"ecoinvent elementary flows {use_version}.xml"
            ) as file:
                root = objectify.parse(file).getroot()

        log.debug(f"Installing biosphere {use_version} for chosen version {version}")
        flow_data = bd.utils.recursive_str_to_unicode(
            [extract_flow_data(ds) for ds in root.iterchildren()]
        )

        return flow_data



from peewee import *
from peewee import _ConnectionState

import threading

from activity_browser.mod.patching import patch_superclass, patched
from activity_browser.ui.core.threading import thread_local


@patch_superclass
class _ConnectionLocal(_ConnectionState, threading.local):

    def __init__(self, *args, **kwargs):
        patched[_ConnectionLocal]["__init__"](self, *args, **kwargs)

        if not hasattr(thread_local, 'peewee_connections'):
            thread_local.peewee_connections = []

        thread_local.peewee_connections.append(self)



from activity_browser.actions.base import ABAction, exception_dialogs
from bw2data import get_node, Node
from activity_browser.ui.icons import qicons
from activity_browser import bwutils

from .activity_redo_allocation import MultifunctionalProcessRedoAllocation


class ActivityModify(ABAction):
    """
    ABAction to delete one or multiple activities if supplied by activity keys. Will check if an activity has any
    downstream processes and ask the user whether they want to continue if so. Exchanges from any downstream processes
    will be removed
    """

    icon = qicons.edit
    text = "Modify Activity"

    @staticmethod
    @exception_dialogs
    def run(activity: tuple | int | Node, field: str, value: any):
        activity = bwutils.refresh_node(activity)

        if field == "product":
            # for some reason product needs to be set like this
            field = "reference product"

        activity[field] = value
        activity.save()



from logging import getLogger

from activity_browser import bwutils
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.mod import bw2data as bd

log = getLogger(__name__)


class CSAddFunctionalUnit(ABAction):
    text = "Add Functional Unit to Calculation Setup"

    @staticmethod
    @exception_dialogs
    def run(cs_name: str, activities: list[tuple | int | bd.Node]):
        activities = [bwutils.refresh_node(node) for node in activities]
        calculation_setup = bd.calculation_setups[cs_name]

        fus = [{act.key: -1.0 if act.get("type") == "waste" else 1.0} for act in activities]
        calculation_setup['inv'] += fus

        bd.calculation_setups[cs_name] = calculation_setup
        bd.calculation_setups.serialize()



from logging import getLogger

from qtpy import QtWidgets

from activity_browser import application
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.mod import bw2data as bd
from activity_browser.ui.icons import qicons
from activity_browser.ui import widgets

from .method_open import MethodOpen

log = getLogger(__name__)


class MethodNew(ABAction):
    """
    ABAction to create a new, empty impact category and open it in edit mode.
    
    This action prompts the user for a new method name using a list edit dialog,
    validates the input, creates an empty method in Brightway2, and opens it
    in the ImpactCategoryDetails page in edit mode so the user can start adding
    characterization factors.
    
    Steps:
    - Open a dialog to prompt the user for the new method name (as a tuple).
    - Validate the new name to ensure it is not empty and does not already exist.
    - Create and register a new empty method in Brightway2.
    - Open the method in the ImpactCategoryDetails page.
    - Set the page to edit mode so the user can add characterization factors.
    """

    icon = qicons.add
    text = "New impact category"

    @staticmethod
    @exception_dialogs
    def run():
        # Open dialog to get new method name
        dialog = widgets.ABListEditDialog(("New Impact Category",), parent=application.main_window)
        dialog.setWindowTitle("New Impact Category")
        
        if dialog.exec_() != QtWidgets.QDialog.Accepted:
            return

        new_name = dialog.get_data(as_tuple=True)

        # Validate new name
        if len(new_name) == 0:
            QtWidgets.QMessageBox.warning(
                application.main_window,
                "Invalid Name",
                "Impact category name cannot be empty.",
            )
            return

        if new_name in bd.methods:
            QtWidgets.QMessageBox.warning(
                application.main_window,
                "Name Already Exists",
                f"An impact category with the name '{' | '.join(new_name)}' already exists.",
            )
            return

        # Create new empty method
        method = bd.Method(new_name)
        method.register()
        method.write([])  # Write empty list of characterization factors
        
        log.info(f"Created new impact category: {new_name}")

        # Open the method in the ImpactCategoryDetails page
        from activity_browser.layouts import pages
        
        page = pages.ImpactCategoryDetailsPage(new_name)
        central = application.main_window.centralWidget()
        central.addToGroup("Characterization Factors", page)
        
        # Set the page to edit mode
        page.is_editable = True
        page.sync()



# -*- coding: utf-8 -*-
import time
from logging import getLogger

import pandas as pd

from activity_browser.mod import bw2data as bd

log = getLogger(__name__)

# Different kinds of indexes, to allow for quick selection of data from
# the Superstructure DataFrame.
# TODO review if this can be made a list instead of pd.Index
SUPERSTRUCTURE = pd.Index(
    [
        "from activity name",
        "from reference product",
        "from location",
        "from categories",
        "from database",
        "from key",
        "to activity name",
        "to reference product",
        "to location",
        "to categories",
        "to database",
        "to key",
        "flow type",
    ]
)


def edit_superstructure_for_string(
    superstructure=SUPERSTRUCTURE, sep="<br>", fhighlight=""
):
    """
    Produces a string format for the essential columns for the scenario difference files with html
    style formatting. Allows for different defined structures.

    Parameters
    ----------
    superstructure: the list of superstructure column headers (by default set to the SUPERSTRUCTURE index,
    this needs to have a defined __str__ operator
    sep: a short string that defines the separator for the column headers, by default this is the html line
    break <br>
    fhighlight: this is provided as a means to highlight the fields, by default this is empty (SHOULD NOT BE
    SET TO None), but could be set to "[]", where the first and last elements enclose the field

    Returns
    -------
    A formatted strign with the required file fields
    """
    text_list = ""
    for field in superstructure:
        text_list += (
            f"{fhighlight[0]}{field}{fhighlight[-1]}{sep}"
            if fhighlight
            else f"{field}{sep}"
        )
    return text_list


def guess_flow_type(row: pd.Series) -> str:
    """Given a series of input- and output keys, make a guess on the flow type."""
    if row.iat[0][0] == bd.config.biosphere:
        return "biosphere"
    elif row.iat[0] == row.iat[1]:
        return "production"
    else:
        return "technosphere"


def _time_it_(func):
    # TODO rename to non_protected name
    """
    For use as a wrapper to time the execution of functions using the python time library
    """

    def wrapper(*args):
        now = time.time()
        result = func(*args)
        log.info(f"{func} -- {time.time() - now}")
        return result

    return wrapper


def parameters_to_sdf(parameter_scenarios: pd.DataFrame) -> pd.DataFrame:
    from activity_browser.bwutils.utils import Parameters
    from activity_browser.bwutils import manager, superstructure

    parameter_scenarios.set_index(["Group", "Name"], inplace=True)

    defaults = [p[:3] for p in Parameters.from_bw_parameters()]
    df = pd.DataFrame(defaults, columns=["Name", "Group", "default"]).set_index(["Group", "Name"])

    for name in [n for n in parameter_scenarios.columns]:
        df[name] = df["default"]

    df.update(parameter_scenarios)

    pm = manager.ParameterManager()
    exchanges = pm.exchanges_from_scenarios(df.columns, [df[scenario] for scenario in df.columns])
    return superstructure.superstructure_from_scenario_exchanges(exchanges)




from typing import List

from qtpy import QtWidgets

import bw2data as bd
import bw_functional as bf

from activity_browser import application
from activity_browser.bwutils import refresh_node
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.ui.icons import qicons
from bw_functional import Product

from .activity_open import ActivityOpen


class ActivityDuplicateToDB(ABAction):
    icon = qicons.duplicate_to_other_database
    text = "Duplicate to other database"

    @classmethod
    @exception_dialogs
    def run(cls, nodes: List[tuple | int | bd.Node], to_db_name: str = None):
        nodes = [refresh_node(node) for node in nodes]
        dbs = {node.get("database") for node in nodes}
        from_db_name = next(iter(dbs))
        from_db_backend = bd.databases[from_db_name]["backend"]

        if not len(dbs) == 1:
            raise ValueError("All selected activities must be from the same database.")

        if any([isinstance(node, bf.Product) for node in nodes]):
            raise ValueError("Products cannot be duplicated to another database. Duplicate the parent process instead.")

        if to_db_name:
            if not cls.confirm_db(to_db_name):
                return
        else:
            to_db_name = cls.request_db(from_db_name)

        to_db_backend = bd.databases[to_db_name]["backend"]

        if from_db_backend == to_db_backend:
            new_nodes = cls.duplicate_simple(nodes, to_db_name)
        elif from_db_backend == "sqlite" and to_db_backend == "functional_sqlite":
            new_nodes = cls.duplicate_sqlite_to_functional_sqlite(nodes, to_db_name)
        elif from_db_backend == "functional_sqlite" and to_db_backend == "sqlite":
            new_nodes = cls.duplicate_functional_sqlite_to_sqlite(nodes, to_db_name)
        else:
            raise NotImplementedError(f"Moving from {from_db_backend} to {to_db_backend} is not supported.")

        ActivityOpen.run(new_nodes)

    @staticmethod
    def request_db(from_db_name: str) -> str | None:
        # get valid databases (not the original database, or locked databases)
        target_dbs = [
            db_name for db_name, meta in bd.databases.items() if
            db_name != from_db_name
            and meta.get("read_only", True) is not True
        ]

        # return if there are no valid databases to duplicate to
        if not target_dbs:
            QtWidgets.QMessageBox.warning(
                application.main_window,
                "No target database",
                "No valid target databases available. Create a new database or set one to writable (not read-only).",
            )
            return

        # construct a dialog where the user can choose a database to duplicate to
        target_db, ok = QtWidgets.QInputDialog.getItem(
            application.main_window,
            "Move node to database",
            "Target database:",
            target_dbs,
            0,
            False,
        )

        return target_db if ok else None

    @staticmethod
    def confirm_db(to_db_name: str):
        user_choice = QtWidgets.QMessageBox.question(
            application.main_window,
            "Move to new database",
            f"Move to {to_db_name} and open as new tab?",
        )
        return user_choice == user_choice.Yes

    @staticmethod
    def duplicate_simple(nodes: list[bd.Node], to_db_name: str) -> list[bd.Node]:
        new_nodes = []

        # move all supplied nodes to the db by copying them
        for node in nodes:
            new_node = node.copy(database=to_db_name)
            new_nodes.append(new_node)

        return new_nodes

    @staticmethod
    def duplicate_sqlite_to_functional_sqlite(nodes: list[bd.Node], to_db_name: str) -> list[bd.Node]:
        from bw_functional.convert import SQLiteToFunctionalSQLite
        new_nodes = []

        for node in nodes:
            dataset = node.as_dict()

            dataset.pop("id", None)
            dataset.pop("key", None)

            dataset["exchanges"] = [exc.as_dict() for exc in node.exchanges()]
            dataset["database"] = to_db_name  # because we didn't copy the dict this will also be reflected in node.key

            new_datasets = SQLiteToFunctionalSQLite.convert_process(node.key, dataset, False)
            new_exchanges = [x for ds in new_datasets.values() for x in ds.pop("exchanges", [])]

            for key, new_dataset in new_datasets.items():
                new_node = bd.Node(**new_dataset)
                new_node.save()
                new_nodes.append(new_node)

            for exc in new_exchanges:
                exc["output"] = (to_db_name, exc["output"][1])  # relink output to new db
                new_exc = bd.Edge(**exc)
                new_exc.save()

        return new_nodes

    @staticmethod
    def duplicate_functional_sqlite_to_sqlite(nodes: list[bd.Node], to_db_name: str) -> list[bd.Node]:
        new_nodes = []
        products = [prod for node in nodes if isinstance(node, bf.Process) for prod in node.products()]

        for product in products:
            dataset = product.as_dict()
            dataset.pop("id", None)
            dataset.pop("key", None)

            exchanges = product.virtual_edges
            dataset["database"] = to_db_name
            dataset["type"] = "processwithreferenceproduct"

            new_node = bd.Node(**dataset)
            new_node.save()
            new_nodes.append(new_node)

            for exc in exchanges:
                exc["output"] = (to_db_name, exc["output"][1])
                if exc["type"] == "production":
                    exc["input"] = (to_db_name, new_node.key[1])
                new_exc = bd.Edge(**exc)
                new_exc.save()

        return new_nodes



from logging import getLogger

from bw2io.ecoinvent import *

import pyprind

from activity_browser.mod.ecoinvent_interface.release import ABEcoinventRelease
from activity_browser.mod.bw2io.importers.ecospold2_biosphere import ABEcospold2BiosphereImporter

log = getLogger(__name__)


def ab_import_ecoinvent_release(version, system_model):
    """Activity Browser version of bw2io.ecoinvent.import_ecoinvent_release that employs a progress slot"""
    from bw2io import migrations
    from .migrations import ab_create_core_migrations

    if not len(migrations):
        ab_create_core_migrations()

    # downloading a release through a AB version of ecoinvent_interface that implements a progress_slot
    release = ABEcoinventRelease(ei.Settings())

    lci_path = release.get_release(
        version=version,
        system_model=system_model,
        release_type=ei.ReleaseType.ecospold,
    )

    # importing biosphere through a biosphere importer that implements a progress_slot
    bio_import = ABEcospold2BiosphereImporter(
        name="biosphere3",
        filepath=lci_path / "MasterData" / "ElementaryExchanges.xml",
    )
    log.info("Applying strategies")
    bio_import.apply_strategies()
    log.info("Writing biosphere database")
    bio_import.write_database()
    bd.preferences["biosphere_database"] = "biosphere3"

    # importing ecoinvent through a ecospold2 importer that implements a progress_slot
    log.info("Importing ecoinvent")
    db_name = f"ecoinvent-{version}-{system_model}"
    ei_import = SingleOutputEcospold2Importer(
        dirpath=str(lci_path / "datasets"),
        db_name=db_name,
        biosphere_database_name="biosphere3",
    )
    log.info("Applying strategies")
    ei_import.apply_strategies()
    log.info("Writing ecoinvent database")
    ei_import.write_database()

    # importing all LCIA methods
    log.info("Gathering LCIA methods")
    lcia_file = ei.get_excel_lcia_file_for_version(release=release, version=version)
    sheet_names = get_excel_sheet_names(lcia_file)

    if "units" in sheet_names:
        units_sheetname = "units"
    elif "Indicators" in sheet_names:
        units_sheetname = "Indicators"
    else:
        raise ValueError(
            f"Can't find worksheet for impact category units in {sheet_names}"
        )

    if "CFs" not in sheet_names:
        raise ValueError(
            f"Can't find worksheet for characterization factors; expected `CFs`, found {sheet_names}"
        )
    log.info("Extracting LCIA methods")
    data = dict(ExcelExtractor.extract(lcia_file))
    units = header_dict(data[units_sheetname])

    log.info("Mapping LCIA methods")
    cfs = header_dict(data["CFs"])

    CF_COLUMN_LABELS = {
        "3.4": "cf 3.4",
        "3.5": "cf 3.5",
        "3.6": "cf 3.6",
    }
    cf_col_label = CF_COLUMN_LABELS.get(version, "cf")
    units_col_label = pick_a_unit_label_already(units[0])
    units_mapping = {
        (row["method"], row["category"], row["indicator"]): row[units_col_label]
        for row in units
    }

    biosphere_mapping = {}
    for flow in bd.Database("biosphere3"):
        biosphere_mapping[(flow["name"],) + tuple(flow["categories"])] = flow.key
        if flow["name"].startswith("[Deleted]"):
            biosphere_mapping[
                (flow["name"].replace("[Deleted]", ""),) + tuple(flow["categories"])
                ] = flow.key

    lcia_data_as_dict = defaultdict(list)

    unmatched = set()
    substituted = set()

    for row in cfs:
        impact_category = (row["method"], row["category"], row["indicator"])
        if row[cf_col_label] is None:
            continue
        try:
            lcia_data_as_dict[impact_category].append(
                (
                    biosphere_mapping[
                        drop_unspecified(
                            row["name"], row["compartment"], row["subcompartment"]
                        )
                    ],
                    float(row[cf_col_label]),
                )
            )
        except KeyError:
            # How is this possible? We are matching ecoinvent data against
            # ecoinvent data from the same release! And yet it moves...
            category = (
                (row["compartment"], row["subcompartment"])
                if row["subcompartment"].lower() != "unspecified"
                else (row["compartment"],)
            )
            same_context = {
                k[0]: v for k, v in biosphere_mapping.items() if k[1:] == category
            }
            candidates = sorted(
                [
                    (damerau_levenshtein(name, row["name"]), name)
                    for name in same_context
                ]
            )
            if (
                    candidates[0][0] < 3
                    and candidates[0][0] != candidates[1][0]
                    and candidates[0][1][0].lower() == row["name"][0].lower()
            ):
                new_name = candidates[0][1]
                pair = (new_name, row["name"])
                if pair not in substituted:
                    print(f"Substituting {new_name} for {row['name']}")
                    substituted.add(pair)
                lcia_data_as_dict[impact_category].append(
                    (
                        same_context[new_name],
                        float(row[cf_col_label]),
                    )
                )
            else:
                if row["name"] not in unmatched:
                    print(
                        "Skipping unmatched flow {}:({}, {})".format(
                            row["name"], row["compartment"], row["subcompartment"]
                        )
                    )
                    unmatched.add(row["name"])

    from activity_browser import signals
    from bw2data import methods
    signals.meta.blockSignals(True)
    signals.method.blockSignals(True)
    old = methods.data.copy()

    for key in pyprind.prog_bar(lcia_data_as_dict, title="Writing LCIA methods"):
        method = bd.Method(key)
        method.register(
            unit=units_mapping.get(key, "Unknown"),
            filepath=str(lcia_file),
            ecoinvent_version=version,
            database="biosphere3",
        )
        method.write(lcia_data_as_dict[key])

    signals.meta.blockSignals(False)
    signals.method.blockSignals(False)
    signals.meta.methods_changed.emit(old, methods.data)



from typing import Any
from urllib.parse import urljoin
from logging import getLogger

from qtpy import QtWidgets, QtCore

from bw2io import install_project
import requests

from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.mod import bw2data as bd
from activity_browser.ui import icons, widgets

log = getLogger(__name__)


class CatalogueModel(QtCore.QAbstractTableModel):
    def __init__(self):
        super().__init__()
        self._data = []
        self._sorted = [key for key in self._data]

    def populate(self, data: dict) -> None:
        self._data = data
        self._sorted = [key for key in self._data]

    def data(self, index: int, role: int):
        if role == QtCore.Qt.DisplayRole:
            return self._sorted[index.row()]
        elif role == QtCore.Qt.ToolTipRole:
            return self._data[self._sorted[index.row()]]

    def rowCount(self, index: int) -> int:
        return len(self._data)

    def columnCount(self, index: int) -> int:
        return 1

    def headerData(self, section:int, orientation:QtCore.Qt.Orientation, role: int=QtCore.Qt.DisplayRole) -> Any:
        if orientation == QtCore.Qt.Horizontal and role == QtCore.Qt.DisplayRole:
            return "Available projects"
        return None


class CatalogueTable(QtWidgets.QTableView):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setVerticalScrollMode(QtWidgets.QTableView.ScrollPerPixel)
        self.setHorizontalScrollMode(QtWidgets.QTableView.ScrollPerPixel)
        self.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)

        self.setWordWrap(True)
        self.setAlternatingRowColors(True)
        self.setSortingEnabled(False)

        self.model = CatalogueModel()
        self.setModel(self.model)

        self.horizontalHeader().setStretchLastSection(True)
        self.horizontalHeader().setHighlightSections(False)
        self.horizontalHeader().setDefaultAlignment(QtCore.Qt.AlignLeft)
        self.verticalHeader().setVisible(False)
        self.setTabKeyNavigation(False)
        self.setSelectionMode(QtWidgets.QAbstractItemView.SelectionMode.SingleSelection)

        self.table_name = "Available projects"
        # Make sure the selected projects is still visible after the focus leaves the table
        self.setStyleSheet("QTableView:!active {selection-background-color: lightgray;}")

    def populate(self, data: dict) -> None:
        self.model.populate(data)
        self.model.layoutChanged.emit()


class ProjectRemoteImportWindow(QtWidgets.QDialog):
    def __init__(self):
        super().__init__()
        self.setWindowTitle("Import project from remote server")
        self.setSizePolicy(
            QtWidgets.QSizePolicy(
                QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding
            )
        )

        layout = QtWidgets.QVBoxLayout()
        dialog_spacing = 8

        remote_url_layout = QtWidgets.QHBoxLayout()
        remote_url_layout.setAlignment(QtCore.Qt.AlignLeft)
        remote_url_layout.addWidget(widgets.ABLabel.demiBold("Remote URL:"))
        self.remote_url_path = QtWidgets.QLineEdit()
        self.remote_url_path.setText("https://files.brightway.dev/")
        self.remote_url_path.textChanged.connect(self._handle_url_changed)
        remote_url_layout.addWidget(self.remote_url_path)
        layout.addLayout(remote_url_layout)

        remote_catalogue_layout = QtWidgets.QHBoxLayout()
        remote_catalogue_layout.setAlignment(QtCore.Qt.AlignLeft)
        remote_catalogue_layout.addWidget(widgets.ABLabel.demiBold("Catalogue file:"))
        self.remote_catalogue = QtWidgets.QLineEdit()
        self.remote_catalogue.setText("projects-config.json")
        self.remote_catalogue.textChanged.connect(self._handle_url_changed)
        remote_catalogue_layout.addWidget(self.remote_catalogue)
        layout.addLayout(remote_catalogue_layout)
        layout.addSpacing(dialog_spacing)

        refresh_button_layout = QtWidgets.QHBoxLayout()
        self.refresh_button = QtWidgets.QPushButton("Download catalogue")
        refresh_button_layout.addWidget(self.refresh_button)
        self.refresh_button.clicked.connect(self._populate_table)
        layout.addLayout(refresh_button_layout)
        layout.addSpacing(dialog_spacing)

        self.table = CatalogueTable()
        self.table.selectionModel().selectionChanged.connect(
            self._handle_table_selection_changed
        )
        layout.addWidget(self.table)
        layout.addSpacing(dialog_spacing)

        project_name_layout = QtWidgets.QHBoxLayout()
        project_name_layout.setAlignment(QtCore.Qt.AlignLeft)
        project_name_layout.addWidget(widgets.ABLabel.demiBold("Project name:"))
        self.project_name = QtWidgets.QLineEdit()
        self.project_name.setText("")
        self.project_name.textChanged.connect(self._handle_project_name_changed)
        project_name_layout.addWidget(self.project_name)
        layout.addLayout(project_name_layout)

        self._overwrite_checkbox = QtWidgets.QCheckBox("Overwrite existing project")
        self._overwrite_checkbox.clicked.connect(self._handle_overwrite_clicked)
        layout.addWidget(self._overwrite_checkbox)

        self._activate_project_checkbox = QtWidgets.QCheckBox("Activate project after import")
        self._activate_project_checkbox.setChecked(True)
        layout.addWidget(self._activate_project_checkbox)

        import_button_layout = QtWidgets.QHBoxLayout()
        self.import_button = QtWidgets.QPushButton("Create project")
        import_button_layout.addWidget(self.import_button)
        self.import_button.clicked.connect(self._import_project)
        layout.addLayout(import_button_layout)
        self._message_label = QtWidgets.QLabel("")
        layout.addWidget(self._message_label)

        self.setLayout(layout)
        self._last_url = ""
        # Initialize the dialog
        self._populate_table()

    def _reset_dialog(self):
        self.table.setEnabled(False)
        self.table.populate(dict())
        self.table.selectionModel().clearSelection()
        self.project_name.setEnabled(False)
        self.project_name.setPlaceholderText("")
        self._overwrite_checkbox.setEnabled(False)
        self._overwrite_checkbox.setChecked(False)
        self._activate_project_checkbox.setEnabled(False)
        self.import_button.setEnabled(False)
        self._message_label.setText("")

    def url(self) -> str:
        return urljoin(self.remote_url_path.text(), self.remote_catalogue.text())

    def _populate_table(self):
        self._reset_dialog()
        try:
            self.refresh_button.setText("Downloading...")
            self.refresh_button.setEnabled(False)
            self.repaint()
            self.setCursor(QtCore.Qt.WaitCursor)
            self._last_url = self.url()
            data = requests.get(self._last_url).json()
            self.table.setEnabled(True)
            self.project_name.setEnabled(True)
            self._activate_project_checkbox.setEnabled(True)
            success = True
        except:
            data = {"Error loading catalogue": None}
            self._message_label.setText("Load a valid catalogue")
            success = False

        self.refresh_button.setText("Download catalogue")
        self.refresh_button.setEnabled(True)
        self.setCursor(QtCore.Qt.ArrowCursor)
        self.table.populate(data)
        if success:
            self._check_project_already_exists()

    def _selected_project_name(self) -> str:
        """Return the selected project name."""
        selection = self.table.selectedIndexes()
        if selection:
            selected_item: QtCore.QModelIndex = selection[0]
            if selected_item.isValid():
                return selected_item.data()
        return ""

    def _project_name(self) -> str:
        """Return the user typed project name or, if empty, the selected one."""
        if self.project_name.text() == "":
            return self._selected_project_name()
        return self.project_name.text()

    def _handle_url_changed(self):
        if self._last_url != self.url():
            self._reset_dialog()
            self._message_label.setText("Load a valid catalogue")

    def _handle_table_selection_changed(self):
        """
        Update the UI when the table selection changes.

        We set the currently selected project name as placeholder text,
        to hint that it can be changed, or will be used as default.
        """
        self.project_name.setPlaceholderText(self._selected_project_name())
        self._check_project_already_exists()

    def _handle_project_name_changed(self):
        self._check_project_already_exists()

    def _unique_project_selection_update(self, selection_valid: bool):
        """
        Update the UI when the selection in the table changes and the
        project name is unique.
        """
        if selection_valid:
            self.import_button.setEnabled(True)
            self._message_label.setText("")
        else:
            self.import_button.setEnabled(False)
            self._message_label.setText("Select a project to import")

    def _duplicate_project_checkbox_update(self):
        """
        Update the UI when the overwrite checkbox state changes and the
        project name is not unique.

        Use the actual state of the checkbox, because it is not
        called only from the checkbox click event.
        """
        if self._overwrite_checkbox.isChecked():
            self.import_button.setEnabled(True)
            self._message_label.setText("")
        else:
            self.import_button.setEnabled(False)
            self._message_label.setText("Project name already exists")

    def _handle_overwrite_clicked(self):
        self._duplicate_project_checkbox_update()

    def _check_project_already_exists(self):
        """
        Update the overwrite checkbox and import button based on the project name.

        If the project already exists, it can only be imported with the
        overwrite flag set. To make sure the user does not import it accidentaly,
        the flag is reset every time the selected project or the project name changes.
        """
        self._overwrite_checkbox.setChecked(False)
        if self._project_name() in bd.projects:
            self._overwrite_checkbox.setEnabled(True)
            self._duplicate_project_checkbox_update()
        else:
            self._overwrite_checkbox.setEnabled(False)
            # Disable the import if there is no selection
            self._unique_project_selection_update(len(self.table.selectedIndexes()) > 0)

    def _import_project(self):
        """
        Import the selected project with the new name.
        It is checked with the UI flow, that there is a catalogue loaded,
        a project to import selected and a unique name provided or the overwrite
        flag is set.
        """
        original_name = self._selected_project_name()
        new_name = self._project_name()
        if original_name and new_name:
            log.info(f"Importing project with name {new_name} "
                        f"(original name {original_name})")
            self.import_button.setText("Creating project...")
            self.import_button.setEnabled(False)
            self.repaint()
            self.setCursor(QtCore.Qt.WaitCursor)

            install_project(
                original_name,
                new_name,
                url=self.remote_url_path.text(),
                overwrite_existing=self._overwrite_checkbox.isChecked()
            )
            if self._activate_project_checkbox.isChecked():
                bd.projects.set_current(new_name, update=False)
            self.setCursor(QtCore.Qt.ArrowCursor)
            self.accept()
        else:
            log.error(f"Project name ({new_name}) or import name ({original_name}) is not valid.")



class ProjectRemoteImport(ABAction):
    """
    ABAction to download a project file from a remote server.
    Allows for customization of server URL, created project name, and whether or not to overwrite existing projects.
    """

    icon = icons.qicons.import_db
    text = "Import remote project"
    tool_tip = "Import a project file from a remote server"

    @staticmethod
    @exception_dialogs
    def run():
        window = ProjectRemoteImportWindow()
        window.adjustSize()
        window.exec_()



from typing import List
from logging import getLogger

from qtpy import QtWidgets

import bw2data as bd

from activity_browser import application, signals
from activity_browser.ui import widgets
from activity_browser.actions.base import ABAction, exception_dialogs

log = getLogger(__name__)


class MethodRename(ABAction):
    """
    Rename an existing impact assessment method (impact category).

    Flow:
    - Ensure only one method is selected and it exists.
    - Prompt the user for the new name and validate it.
    - Copy the method to the new name and process it.
    - Update all Brightway calculation setups by replacing the old method in 'ia' lists with
      the new name and serialize the updates.
    - Emit a rename signal and deregister the old method.

    Raises:
    - ValueError: If more than one method is provided for renaming.
    - RuntimeError: If the method does not exist, the new name is empty, or the new name already exists.
    """

    text = "Rename Impact Category"

    @staticmethod
    @exception_dialogs
    def run(method_name: tuple[str] | list[tuple[str]]):
        # safeguard: only allow renaming one method at a time
        if isinstance(method_name, list):
            if len(method_name) != 1 or not isinstance(method_name[0], tuple):
                raise ValueError("Can only rename one method at a time.")
            method_name = method_name[0]

        # check if method exists
        if method_name not in bd.methods:
            raise RuntimeError(f"Method {method_name} does not exist.")

        method = bd.Method(method_name)

        # open dialog to get new name
        dialog = widgets.ABListEditDialog(
            method_name,
            title="Rename Impact Category",
            parent=application.main_window,
        )

        # execute the dialog and check for acceptance
        if dialog.exec_() != QtWidgets.QDialog.Accepted:
            return

        new_name = dialog.get_data(as_tuple=True)

        # check new name validity
        if new_name == method_name:
            return  # no change

        if len(new_name) == 0:
            raise RuntimeError("Method name cannot be empty.")

        if new_name in bd.methods:
            raise RuntimeError(f"Method {new_name} already exists.")

        # copy method to new name and process
        method.copy(new_name).process()

        # Update any calculation setups that reference this method
        MethodRename.rename_method_in_calculation_setups(method_name, new_name)

        # this should not happen like this, as the model and therefore signals should be handled declaritavely,
        # but since method renaming is not native to bw2data we have to do it manually here
        signals.method.renamed.emit(method_name, new_name)

        # deregister old method
        method.deregister()

    @staticmethod
    def rename_method_in_calculation_setups(old_name: tuple, new_name: tuple) -> None:
        """Replace occurrences of old_name with new_name in all CS 'ia' lists and serialize.

        Handles both tuple and single-string method keys. Best-effort: logs on failure
        without blocking the rename flow.
        """
        try:
            changed_any = False

            for cs_name, cs in bd.calculation_setups.items():
                ia = cs.get("ia", [])
                
                if old_name not in ia:
                    continue

                i = ia.index(old_name)
                ia[i] = new_name

                changed_any = True
                log.info(
                    f"Updated calculation setup '{cs_name}': renamed impact category {old_name} -> {new_name}"
                )
            
            if changed_any:
                bd.calculation_setups.serialize()
        except Exception:
            log.exception("Failed to update calculation setups after method rename")



from qtpy import QtWidgets

import bw2data as bd

from activity_browser import application
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.mod.bw2io import remote
from activity_browser.ui.icons import qicons
from activity_browser.ui.core.threading import ABThread

from .project_switch import ProjectSwitch


class ProjectNewRemote(ABAction):
    """
    ABAction to create a new project from a remote template.
    """

    icon = qicons.add
    text = "New project from remote"
    tool_tip = "Make a new project from remote template"

    @staticmethod
    @exception_dialogs
    def run(project_key: str):
        name, ok = QtWidgets.QInputDialog.getText(
            application.main_window,
            "Create project from remote",
            "Name of new project:" + " " * 25,
        )

        if not ok or not name:
            return

        if name in bd.projects:
            QtWidgets.QMessageBox.information(
                application.main_window,
                "Not possible.",
                "A project with this name already exists.",
            )
            return

        thread = InstallThread(application)
        thread.start(project_key, name)

        dialog = MigrateDialog(application.main_window)
        dialog.show()

        thread.finished.connect(dialog.close)
        thread.finished.connect(lambda: ProjectSwitch.run(name))


class MigrateDialog(QtWidgets.QProgressDialog):
    def __init__(self, parent):
        super().__init__(parent)
        self.setWindowTitle("Installing project")
        self.setLabelText("Restoring project from template, this may take a while...")
        self.setRange(0, 0)
        self.setCancelButton(None)


class InstallThread(ABThread):
    def run_safely(self, project_key: str, name: str):
        remote.install_project(project_key, name)




import copy

from qtpy import QtWidgets

import bw2data as bd
import bw_functional as bf

from activity_browser import application
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.ui.icons import qicons
from activity_browser.ui.core.threading import ABThread

from .database_new import NewDatabaseDialog


class DatabaseDuplicate(ABAction):
    """
    ABAction to duplicate a database. Asks the user to provide a new name for the database, and returns when the name is
    already in use by an existing database. Then it shows a progress dialogue which will construct a new thread in which
    the database duplication will take place. This thread instructs the DatabaseController to duplicate the selected
    database with the chosen name.
    """

    icon = qicons.duplicate_database
    text = "Duplicate database..."
    tool_tip = "Make a duplicate of this database"

    @staticmethod
    @exception_dialogs
    def run(db_name: str):
        assert db_name in bd.databases
        backend = bd.databases[db_name].get("backend", "undefined")

        if backend not in ["sqlite", "functional_sqlite"]:
            QtWidgets.QMessageBox.information(
                application.main_window,
                "Not possible",
                f"Unsupported database backend {backend}",
            )
            return

        name, backend, ok = NewDatabaseDialog.get_new_database_data(window_title="Duplicate database", backend=backend)

        if not name or not ok:
            return

        if name in bd.databases:
            QtWidgets.QMessageBox.information(
                application.main_window,
                "Not possible",
                "A database with this name already exists.",
            )
            return

        DuplicateDatabaseDialog(db_name, name, backend, application.main_window)


class DuplicateDatabaseDialog(QtWidgets.QProgressDialog):
    def __init__(self, from_db: str, to_db: str, backend: str, parent=None):
        super().__init__(parent=parent)
        self.setWindowTitle("Duplicating database")
        self.setLabelText(
            f"Duplicating existing database <b>{from_db}</b> to new database <b>{to_db}</b>:"
        )
        self.setModal(True)
        self.setRange(0, 0)

        self.dup_thread = DuplicateDatabaseThread(application)
        self.dup_thread.finished.connect(self.thread_finished)

        self.show()

        self.dup_thread.start(from_db, to_db, backend)

    def thread_finished(self) -> None:
        self.dup_thread.exit(0)
        self.setMaximum(1)
        self.setValue(1)


class DuplicateDatabaseThread(ABThread):

    def run_safely(self, copy_from, copy_to, backend):
        database = bd.Database(copy_from)

        data = database.load()
        data = database.relabel_data(data, copy_from, copy_to)

        new_database = bd.Database(copy_to, backend=backend)

        metadata = copy.copy(database.metadata)
        metadata["format"] = f"Copied from '{copy_from}'"
        metadata["backend"] = backend
        new_database.register(**metadata)

        if database.backend == "sqlite" and backend == "functional_sqlite":
            data = bf.convert_sqlite_to_functional_sqlite(data)
        elif database.backend == "functional_sqlite" and backend == "sqlite":
            data = bf.convert_functional_sqlite_to_sqlite(data)

        new_database.write(data, searchable=metadata.get("searchable"))
        return new_database



from abc import abstractmethod
from collections import defaultdict
from collections.abc import Iterator
from typing import Iterable, Optional, Tuple

import numpy as np
from stats_arrays import MCRandomNumberGenerator, UncertaintyBase

from bw2data.backends import ExchangeDataset
from bw2data.parameters import get_new_symbols, ParameterizedExchange
from bw2calc import LCA
from bw2parameters import ParameterSet, MissingName, Interpreter

from .utils import Index, Indices, Parameters, StaticParameters


class ParameterManager(object):
    """A manager for Brightway2 parameters, allowing for formula evaluation
    without touching the database.
    """

    def __init__(self):
        self.parameters: Parameters = Parameters.from_bw_parameters()
        self.initial: StaticParameters = StaticParameters()
        self.indices: Indices = self.construct_indices()

    def construct_indices(self) -> Indices:
        """Given that ParameterizedExchanges will always have the same order of
        indices, construct them once and reuse when needed.
        """
        indices = Indices()
        for p in self.initial.act_by_group_db:
            params = self.initial.exc_by_group(p.group)
            indices.extend(
                Index.build_from_exchange(ExchangeDataset.get_by_id(pk))
                for pk in params
            )
        return indices

    def recalculate_project_parameters(self) -> dict:
        data = self.initial.project()
        if not data:
            return {}

        new_values = self.parameters.data_by_group("project")

        for name, amount in new_values.items():
            data[name]["amount"] = amount

        ParameterSet(data).evaluate_and_set_amount_field()
        return StaticParameters.prune_result_data(data)

    def recalculate_database_parameters(
        self, database: str, global_params: dict = None
    ) -> dict:
        data = self.initial.by_database(database)
        if not data:
            return {}

        glo = global_params or {}
        new_values = self.parameters.data_by_group(database)
        for name, amount in new_values.items():
            data[name]["amount"] = amount

        new_symbols = get_new_symbols(data.values(), set(data))
        missing = new_symbols.difference(glo)
        if missing:
            raise MissingName(
                "The following variables aren't defined:\n{}".format("|".join(missing))
            )

        glo = Parameters.static(glo, needed=new_symbols) if new_symbols else None

        ParameterSet(data, glo).evaluate_and_set_amount_field()
        return StaticParameters.prune_result_data(data)

    def process_database_parameters(self, global_params: dict = None) -> dict:
        glo = global_params or self.recalculate_project_parameters()
        all_db = {}
        for database in self.initial.databases:
            db = self.recalculate_database_parameters(database, glo)
            all_db[database] = {x: y for x, y in db.items()} if db else {}
        return all_db

    def recalculate_activity_parameters(
        self, group: str, global_params: dict = None
    ) -> dict:
        data = self.initial.act_by_group(group)
        if not data:
            return {}

        new_values = self.parameters.data_by_group(group)
        glo = global_params or {}
        for name, amount in new_values.items():
            data[name]["amount"] = amount

        new_symbols = get_new_symbols(data.values(), set(data))
        missing = new_symbols.difference(global_params)
        if missing:
            raise MissingName(
                "The following variables aren't defined:\n{}".format("|".join(missing))
            )

        glo = Parameters.static(glo, needed=new_symbols) if new_symbols else None

        ParameterSet(data, glo).evaluate_and_set_amount_field()
        return StaticParameters.prune_result_data(data)

    def recalculate_exchanges(
        self, group: str, global_params: dict = None
    ) -> Iterable[Tuple[int, float]]:
        """Constructs a list of exc.id/amount tuples for the
        ParameterizedExchanges in the given group.
        """
        params = self.initial.exc_by_group(group)
        if not params:
            return []

        glo = global_params or {}
        interpreter = Interpreter()
        interpreter.symtable.update(glo)
        return [(k, interpreter(v)) for k, v in params.items()]

    def process_exchanges(
        self,
        global_params: dict = None,
        db_params: dict = None,
        build_indices: bool = True,
    ) -> np.ndarray:
        dbs = db_params or {}
        complete_data = np.zeros(len(self.indices))

        offset = 0
        for p in self.initial.act_by_group_db:
            combination = (
                {x: y for x, y in global_params.items()} if global_params else {}
            )
            combination.update(dbs.get(p.database, {}))
            combination.update(
                self.recalculate_activity_parameters(p.group, combination)
            )

            recalculated = self.recalculate_exchanges(
                p.group, global_params=combination
            )
            # If the parameter group contains no ParameterizedExchanges, skip.
            if not recalculated:
                continue
            # `data` contains the recalculated amounts for the exchanges.
            _, data = zip(*recalculated)
            complete_data[offset : len(data) + offset] = data
            offset += len(data)

        return complete_data

    def calculate(self) -> np.ndarray:
        """Convenience function that takes calculates the current parameters
        and returns a fully-formed set of exchange amounts and indices.

        All parameter types are recalculated in turn before interpreting the
        ParameterizedExchange formulas into amounts.
        """
        global_project = self.recalculate_project_parameters()
        all_db = self.process_database_parameters(global_project)
        data = self.process_exchanges(global_project, all_db)
        return data

    @abstractmethod
    def recalculate(self, values: dict[str, float]) -> np.ndarray:
        """Convenience function that takes the given new values and recalculates.
        Returning a fully-formed set of exchange amounts and indices.

        All parameter types are recalculated in turn before interpreting the
        ParameterizedExchange formulas into amounts.
        """
        self.parameters.update(values)
        return self.calculate()

    def ps_recalculate(self, values: dict[str, float]) -> np.ndarray:
        """Used to recalculate brightway parameters without editing the database.
        Leftover from Presamples.

        Side-note on presamples: Presamples was used in AB for calculating scenarios,
        presamples was superseded by this implementation. For more reading:
        https://presamples.readthedocs.io/en/latest/index.html"""
        data = self.recalculate(values)
        # After recalculating all the exchanges format them according to
        # presamples requirements: samples as a column of floats.
        samples = data.reshape(1, -1).T
        return samples

    def reformat_indices(self) -> np.ndarray:
        """Additional information is required for storing the indices as presamples.
        Leftover from Presamples.

        Side-note on presamples: Presamples was used in AB for calculating scenarios,
        presamples was superseded by this implementation. For more reading:
        https://presamples.readthedocs.io/en/latest/index.html"""
        result = np.zeros(len(self.indices), dtype=object)
        for i, idx in enumerate(self.indices):
            result[i] = (idx.input, idx.output, idx.flow_type)
        return result

    def arrays_from_scenarios(self, scenarios) -> (np.ndarray, np.ndarray):
        """Used to generate exchange scenario data from parameter scenario data.
        Leftover from Presamples.

        Side-note on presamples: Presamples was used in AB for calculating scenarios,
        presamples was superseded by this implementation. For more reading:
        https://presamples.readthedocs.io/en/latest/index.html"""
        sample_data = [self.ps_recalculate(values.to_dict()) for _, values in scenarios]
        samples = np.concatenate(sample_data, axis=1)
        indices = self.reformat_indices()
        return samples, indices

    @staticmethod
    def has_parameterized_exchanges() -> bool:
        """Test if ParameterizedExchanges exist, no point to using this manager
        otherwise.
        """
        return ParameterizedExchange.select().exists()

    def parameter_exchange_dependencies(self) -> dict:
        """

        Schema: {param1: List[tuple], param2: List[tuple]}
        """
        parameters = defaultdict(list)
        for act in self.initial.act_by_group_db:
            exchanges = self.initial.exc_by_group(act.group)
            for exc, formula in exchanges.items():
                params = get_new_symbols([formula])
                # Convert exchange from int to Index
                exc = Index.build_from_exchange(ExchangeDataset.get_by_id(exc))
                for param in params:
                    parameters[param].append(exc)
        return parameters

    def extract_active_parameters(self, lca: LCA) -> dict:
        """Given an LCA object, extract the used exchanges and build a
        dictionary of parameters with the exchanges that use these parameters.

        Schema: {param1: {"name":  str, "act": Optional[tuple],
                          "group": str, "exchanges": List[tuple]}}
        """
        params = self.parameter_exchange_dependencies()
        used_exchanges = set(lca.biosphere_dict.keys())
        used_exchanges = used_exchanges.union(lca.activity_dict.keys())

        # Make a pass through the 'params' dictionary and remove exchanges
        # that don't exist in the 'used_exchanges' set.
        for p in params:
            keep = [
                i
                for i, exc in enumerate(params[p])
                if (exc.input in used_exchanges and exc.output in used_exchanges)
            ]
            params[p] = [params[p][i] for i in keep]
        # Now only keep parameters with exchanges.
        keep = [p for p, excs in params.items() if len(excs) > 0]
        schema = {
            p: {"name": p, "act": None, "group": None, "exchanges": params[p]}
            for p in keep
        }
        # Now start filling in the parameters that remain with information.
        for group in self.initial.groups:
            for name, data in self.initial.act_by_group(group).items():
                key = (data.get("database"), data.get("code"))
                if name in schema:
                    # Make sure that the activity parameter matches the
                    # output of the exchange.
                    exc = next(e.output for e in schema[name]["exchanges"])
                    if key == exc:
                        schema[name]["name"] = name
                        schema[name]["group"] = group
                        schema[name]["act"] = key
        # Lastly, determine for the remaining parameters if they are database
        # or project parameters.
        for db in self.initial.databases:
            for name, data in self.initial.by_database(db).items():
                if name in schema and schema[name]["group"] is None:
                    schema[name]["group"] = db
        for name in (n for n in schema if schema[n]["group"] is None):
            schema[name]["group"] = "project"
        return schema

    def new_process_exchanges(self, project_params: dict, database_params: dict):
        complete = {}
        for p in self.initial.act_by_group_db:

            scope = project_params.copy()
            scope.update(database_params.get(p.database, {}))

            activity_params = self.recalculate_activity_parameters(p.group, scope)

            scope.update(activity_params)

            exchanges = {exc_id: value for exc_id, value in self.recalculate_exchanges(p.group, scope)}

            complete.update(exchanges)

        return complete

    def exchanges_from_scenarios(self, scenario_names, data):
        scenario_dict = {name: dict(data[i]) for i, name in enumerate(scenario_names)}
        scenario_samples = {}

        for scenario, parameters in scenario_dict.items():
            self.parameters.update(parameters)

            project_params = self.recalculate_project_parameters()
            database_params = self.process_database_parameters(project_params)
            scenario_samples[scenario] = self.new_process_exchanges(project_params, database_params)
        return scenario_samples


class MonteCarloParameterManager(ParameterManager, Iterator):
    """Use to sample the uncertainty of parameter values, mostly for use in
    Monte Carlo calculations.

    Each iteration will sample the parameter uncertainty, after which
    all parameters and parameterized exchanges are recalculated. These
    recalculated values are then returned as a simplified `params` array,
    which is similar to the `tech_params` and `bio_params` arrays in the
    LCA classes.

    Makes use of the `MCRandomNumberGenerator` to sample from all of the
    distributions in the same way.

    """

    def __init__(self, seed: Optional[int] = None):
        super().__init__()
        self.uncertainties = UncertaintyBase.from_dicts(
            *[p.data for p in self.parameters]
        )
        self.mc_generator = MCRandomNumberGenerator(self.uncertainties, seed=seed)

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def recalculate(self, iterations: int = 10) -> np.ndarray:
        assert iterations > 0, "Must have a positive amount of iterations"
        if iterations == 1:
            return self.next()
        # Construct indices, prepare sized array and sample parameter
        # uncertainty distributions `interations` times.
        all_data = np.empty((iterations, len(self.indices)), dtype=Indices.array_dtype)
        random_bounded_values = self.mc_generator.generate(iterations)

        # Now, repeatedly replace parameter amounts with sampled data and
        # recalculate. Every processed row is added to the sized array.
        for i in range(iterations):
            values = random_bounded_values.take(i, axis=1)
            self.parameters.update(values)
            data = self.calculate()
            all_data[i] = self.indices.mock_params(data)

        return all_data

    def next(self) -> np.ndarray:
        """Similar to `recalculate` but only performs a single sampling and
        recalculation.
        """
        values = self.mc_generator.next()
        keys = [(p.group, p.name) for p in self.parameters]
        self.parameters.update({key: value for key, value in zip(keys, values)})
        data = self.calculate()
        return self.indices.mock_params(data)

    def retrieve_sampled_values(self, data: dict):
        """Enters the sampled values into the 'exchanges' list in the 'data'
        dictionary.
        """
        for name, vals in data.items():
            param = next(
                (
                    p
                    for p in self.parameters
                    if p.name == vals.get("name") and p.group == vals.get("group")
                ),
                None,
            )
            if param is None:
                continue
            data[name]["values"].append(param.amount)



import ast
import os.path
from importlib.metadata import PackageNotFoundError, version
from logging import getLogger

from .utils import safe_link_fetch, sort_semantic_versions

log = getLogger(__name__)

# get AB version
try:
    __version__ = version(__package__)
except PackageNotFoundError:
    __version__ = "0.0.0"


def get_compatible_versions() -> list:
    """Get compatible versions of ecoinvent for this AB version.

    Reads this file on github repo: activity-browser/better_biosphere_handling/compatible_ei_versions.txt'.
    Converts file content to available ecoinvent versions for each version of AB.
    Finds the correct available versions for this AB version, if failing to read version,
    the lowest version in the file is chosen.
    """
    try:
        # read versions
        versions_URL = "https://raw.githubusercontent.com/LCA-ActivityBrowser/activity-browser/main/activity_browser/bwutils/ecoinvent_biosphere_versions/compatible_ei_versions.txt"
        page, error = safe_link_fetch(versions_URL)
        if not error:
            file = page.text
        else:
            # silently try a local fallback:
            log.debug(
                f"Reading online compatible ecoinvent versions failed "
                f"-attempting local fallback- with this error: {error}"
            )
            file_path = os.path.join(
                os.path.dirname(__file__),
                "bwutils",
                "ecoinvent_biosphere_versions",
                "compatible_ei_versions.txt",
            )
            with open(file_path, "r") as f:
                file = f.read()
        all_versions = ast.literal_eval(file)

        # select either the latest lower version available or if none available the lowest version for safety
        sorted_versions = sort_semantic_versions(all_versions.keys())
        for ab_version in sorted_versions:
            if sort_semantic_versions([__version__, ab_version])[0] == __version__:
                # current version is higher than or equal to tested AB version:
                ei_versions = all_versions[ab_version]
                break
        else:
            ei_versions = all_versions[sorted_versions[-1]]

        log.debug(
            f"Following versions of ecoinvent are compatible with AB {__version__}: {ei_versions}"
        )
        return ei_versions

    except Exception as error:
        log.debug(f"Reading local fallback failed with: {error}")
        return ["3.4", "3.5", "3.6", "3.7", "3.7.1", "3.8", "3.9", "3.9.1"]


__ei_versions__ = get_compatible_versions()



import threading
import logging

from qtpy.QtCore import QThread, SignalInstance, Signal
from qtpy import QtWidgets


class ABThread(QThread):
    status: SignalInstance = Signal(int, str)
    exception: SignalInstance = Signal(Exception)

    _args = ()
    _kwargs = {}

    def __init__(self, parent=None):
        super().__init__(parent)

        from activity_browser import application
        self.exception.connect(application.main_window.dialog_on_exception)

    def start(self, *args, priority=QThread.NormalPriority, **kwargs):
        """
        Reimplemented from QThread to set the priority of the thread.
        """
        self._args = args
        self._kwargs = kwargs
        super().start(priority)

    def run(self):
        """Reimplemented from QThread to close any database connections before finishing."""
        from activity_browser.mod.tqdm.std import qt_tqdm

        qt_tqdm.updated.connect(self._emit_status)

        # call run_safely and finish by closing the connections
        with SafeBWConnection(), InfoToSlot(self.status.emit):
            try:
                self.run_safely(*self._args, **self._kwargs)
            except Exception as e:
                # send exception signal
                self.exception.emit(e)
                raise e

        qt_tqdm.updated.disconnect(self._emit_status)
        self.status.emit(100, "Complete")

    def _emit_status(self, progress: int, message: str):
        if progress == 100:
            self.status.emit(-1, "Working...")
        else:
            self.status.emit(progress, message)

    def run_safely(self, *args, **kwargs):
        raise NotImplementedError

    def connect_progress_dialog(self, progress_dialog: QtWidgets.QProgressDialog):
        """
        Connects the status signal to a progress dialog.
        """
        def slot(progress, message):
            if progress == -1:
                progress_dialog.setLabelText(message)
                progress_dialog.setRange(0, 0)
            else:
                progress_dialog.setRange(0, 100)
                progress_dialog.setValue(progress)
                progress_dialog.setLabelText(message or "Working...")

        self.status.connect(slot)


class SafeBWConnection:
    def __enter__(self):
        return

    def __exit__(self, *args):
        """
        Closes all connections for this thread
        """
        for conn in getattr(thread_local, "peewee_connections", []):
            if hasattr(conn, "conn") and hasattr(conn.conn, "close"):
                conn.conn.close()


class InfoToSlot:
    def __init__(self, progress_slot=lambda progress, message: None):
        self.handler = LoggingProgressHandler("INFO")
        thread_local.progress_slot = progress_slot

    def __enter__(self):
        logging.root.addHandler(self.handler)
        return

    def __exit__(self, *args):
        logging.root.removeHandler(self.handler)
        return


class LoggingProgressHandler(logging.Handler):
    def filter(self, record: logging.LogRecord) -> bool:
        if record.thread != threading.get_ident():
            return False
        if record.levelname != "INFO":
            return False
        return True

    def emit(self, record: logging.LogRecord):
        try:
            thread_local.progress_slot(None, record.message)
        except AttributeError:
            pass


thread_local = threading.local()
